# -*- coding: utf-8 -*-
"""
Created on Mon Oct 29 08:39:57 2018

@author: Marcus
"""

import numpy as np

def get_defaults(et_name):

    settings = Settings(et_name)

    if et_name == 'IS4_Large_Peripheral':
        settings.SAMPLING_RATE = 90
    elif et_name == 'Tobii Pro Spectrum':
        settings.SAMPLING_RATE = 600
        settings.TRACKING_MODE = 'human'
    elif et_name == 'Tobii Pro Nano':
        settings.SAMPLING_RATE = 60
    elif et_name == 'Tobii Pro Spark':
        settings.SAMPLING_RATE = 60
    elif et_name == 'Tobii TX300':
        settings.SAMPLING_RATE = 300
    elif et_name == 'Tobii T60 XL':
        settings.SAMPLING_RATE = 60
    elif et_name == 'Tobii Pro X3-120 EPU':
        settings.SAMPLING_RATE = 120
    elif et_name == 'Tobii Pro X3-120':
        settings.SAMPLING_RATE = 120
    elif et_name == 'X2-60_Compact':
        settings.SAMPLING_RATE = 60
    elif et_name == 'X2-30_Compact':
        settings.SAMPLING_RATE = 40
    elif et_name == 'Tobii X60':
        settings.SAMPLING_RATE = 60
    elif et_name == 'Tobii X120':
        settings.SAMPLING_RATE = 120
    elif et_name == 'Tobii T60':
        settings.SAMPLING_RATE = 60
    elif et_name == 'Tobii T120':
        settings.SAMPLING_RATE = 120
    elif et_name == 'Tobii Pro Fusion':
        settings.SAMPLING_RATE = 120
    else:
        raise Exception('eye tracker type not supported')

    return settings


class Connect(object):
    def __init__(self, in_arg='dummy'):
        '''  in_arg can be either string with eye tracker name
        or 'settings' generated by calling (and optionally modifying)
        the output from get_defaults()
        '''

        if isinstance(in_arg, str):
            if 'dummy' in in_arg:
                from titta import Tobii_dummy
                self.__class__ = Tobii_dummy.Connect
                self.__class__.__init__(self)
            else:
                from titta import Tobii
                self.__class__ = Tobii.myTobii
                self.__class__.__init__(self, in_arg)
        else:
            from titta import Tobii
            self.__class__ = Tobii.myTobii
            self.__class__.__init__(self, in_arg)



class Settings(object):

    def __init__(self, et_name):
        ''' Default settings for eye tracker
        '''

        self.graphics = Graphics()

        self.DEBUG = False

        # Default name of et-data file
        self.FILENAME                    = 'test.tsv'

        # Tracking parameters
        self.TRACKER_ADDRESS  = ''           # If none is given, find one on the network
        self.SAMPLING_RATE = 600             # Set sampling rate of tracker
        self.eye_tracker_name = et_name
        self.TRACKING_MODE = 'Default'

        # Parameters for calibration
        self.PACING_INTERVAL = 1.0           # How long to present the dot until samples are collected
        self.AUTO_PACE = 2                   # autoaccept (2), semi autoaccept (1, accept first point, default)
                                        # of accept with space bar (0)

        self.ANIMATE_CALIBRATION = True      # Static or animated calibration dots
        self.RECORD_EYE_IMAGES_DURING_CALIBRATION = False
        self.N_CAL_TARGETS = 5               # Valid: 0, 1, 5, 9, 13

        # List all possible calibration points (in Tobii's coordinate system)
        # (0.0, 0.0) is the upper left corner and (1.0, 1.0) is the lower right corner.

        # Define the 13 point array (reading order)
        self.CAL_TARGETS = np.array([[0.1, 0.1], [0.5, 0.1], [.9,.1],
                           [.3,.3], [.7,.3],
                           [.1,.5], [.5,.5], [.9,.5],
                           [.3,.7], [.7,.7],
                           [.1,.9], [.5,.9], [.9,.9]])


        self.VAL_POS_TOBII = np.array([[0.2, 0.5], [0.5, 0.8], [0.8, 0.5], [0.5, 0.2]])

        # CAL_POS_TOBII = np.array([[0.5, 0.5], [0.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 1.0],
        #                           [0.5, 1.0], [0.5, 0.0], [0.0, 0.5], [1.0, 0.5]])

        # if N_CAL_TARGETS == 0:
        #     CAL_POS_TOBII = []
        # elif N_CAL_TARGETS == 1:
        #     CAL_POS_TOBII = CAL_POS_TOBII[0, :]
        # elif N_CAL_TARGETS == 5:
        #     CAL_POS_TOBII = CAL_POS_TOBII[[0, 1, 2, 3, 4], :]

        # # VAL_POS_TOBII = np.array([[0.2, 0.2], [0.2, 0.8], [0.8, 0.2], [0.8, 0.8]])
        # VAL_POS_TOBII = np.array([[0.2, 0.5], [0.5, 0.8], [0.8, 0.5], [0.5, 0.2]])

        # # Scale the positions so they look good on the screen
        # scaling = 0.7
        # corr = 0.5 - (scaling * 0.5)
        # self.CAL_POS_TOBII = CAL_POS_TOBII * scaling + corr
        # self.VAL_POS_TOBII = VAL_POS_TOBII * scaling + corr

class Graphics(object):

    def __init__(self):
        ''' Default settings for graphics
        '''

        blue = tuple(np.array([37, 97, 163]) / 255.0 * 2 - 1)
        blue_active = tuple(np.array([11, 122, 244]) / 255.0 * 2 - 1)
        green = tuple(np.array([0, 120, 0]) / 255.0 * 2 - 1)
        red = tuple(np.array([150, 0, 0]) / 255.0 * 2 - 1)
        yellow = tuple(np.array([255, 255, 0]) / 255.0 * 2 - 1)
        yellow_linecolor = tuple(np.array([255, 255, 0]) / 255.0 * 2 - 1)

        self.blue = blue
        self.blue_active = blue_active

        self.TEXT_SIZE = 0.04 # Size of text (norm)
        self.TEXT_COLOR = 'white'

        self.ET_SAMPLE_RADIUS = 5 # in pix

        # SIze of calibration dots
        self.TARGET_SIZE=30 # in pix
        self.TARGET_SIZE_INNER=self.TARGET_SIZE / 6  # inner diameter of dot

        # Theses parameters are changed directly in the EThead class
        # self.HEAD_POS_CIRCLE_FIXED_COLOR = blue
        # self.HEAD_POS_CIRCLE_FIXED_RADIUS = 0.25

        # self.HEAD_POS_CIRCLE_MOVING_COLOR = yellow
        # self.HEAD_POS_CIRCLE_MOVING_FILLCOLOR = yellow
        # self.HEAD_POS_CIRCLE_MOVING_RADIUS = 0.25
        # self.HEAD_POS_CIRCLE_MOVING_MIN_RADIUS = 0.05

        self.POS_CAL_BUTTON = (0.5, -0.8)
        self.COLOR_CAL_BUTTON =  green
        self.WIDTH_CAL_BUTTON = 0.30
        self.HEIGHT_CAL_BUTTON = 0.08
        self.CAL_BUTTON = 'space'
        self.CAL_BUTTON_TEXT = 'calibrate (spacebar)'

        self.POS_RECAL_BUTTON = (-0.5, -0.8)
        self.COLOR_RECAL_BUTTON =  red
        self.WIDTH_RECAL_BUTTON = 0.30
        self.HEIGHT_RECAL_BUTTON = 0.08
        self.RECAL_BUTTON = 'c'
        self.RECAL_BUTTON_TEXT = 're-calibrate (c)'

        self.POS_REVAL_BUTTON = (-0.21, -0.8)
        self.COLOR_REVAL_BUTTON =  red
        self.WIDTH_REVAL_BUTTON = 0.30
        self.HEIGHT_REVAL_BUTTON = 0.08
        self.REVAL_BUTTON = 'v'
        self.REVAL_BUTTON_TEXT = 're-validate (v)'

        # Button for showing eye images
        self.POS_SETUP_BUTTON = (-0.5, -0.8)
        self.COLOR_SETUP_BUTTON = blue
        self.WIDTH_SETUP_BUTTON = 0.30
        self.HEIGHT_SETUP_BUTTON = 0.08
        self.SETUP_BUTTON = 'e'
        self.SETUP_BUTTON_TEXT = 'eye images (e)'

        self.POS_ACCEPT_BUTTON = (0.5, -0.8)
        self.COLOR_ACCEPT_BUTTON = green
        self.WIDTH_ACCEPT_BUTTON = 0.30
        self.HEIGHT_ACCEPT_BUTTON = 0.08
        self.ACCEPT_BUTTON = 'space'
        self.ACCEPT_BUTTON_TEXT = 'accept (spacebar)'

        self.POS_BACK_BUTTON = (-0.5, -0.8)
        self.COLOR_BACK_BUTTON = blue
        self.WIDTH_BACK_BUTTON = 0.30
        self.HEIGHT_BACK_BUTTON = 0.08
        self.BACK_BUTTON = 'b'
        self.BACK_BUTTON_TEXT = 'basic (b)'

        self.POS_GAZE_BUTTON = (0.8, 0.8)
        self.COLOR_GAZE_BUTTON = blue
        self.WIDTH_GAZE_BUTTON = 0.25
        self.HEIGHT_GAZE_BUTTON = 0.08
        self.GAZE_BUTTON = 'g'
        self.GAZE_BUTTON_TEXT = 'show gaze (g)'

        self.POS_CAL_IMAGE_BUTTON = (-0.8, 0.8)
        self.COLOR_CAL_IMAGE_BUTTON = (0.2, 0.2, 0.2)
        self.WIDTH_CAL_IMAGE_BUTTON = 0.25
        self.HEIGHT_CAL_IMAGE_BUTTON = 0.08
        self.CAL_IMAGE_BUTTON = 's'
        self.CAL_IMAGE_BUTTON_TEXT = 'Show calibration (s)'

        self.SETUP_DOT_OUTER_DIAMETER = 0.03 # Height unit
        self.SETUP_DOT_INNER_DIAMETER = 0.005

        # Parameters for eye images (default values are for Spectrum)
        self.EYE_IMAGE_SIZE = (0.5, 0.25)
        # self.EYE_IMAGE_SIZE_PIX = (175, 496)
        # self.EYE_IMAGE_SIZE_PIX_FULL_FRAME = (512, 640)
        self.EYE_IMAGE_POS_L = (-0.5, -0.4)
        self.EYE_IMAGE_POS_R = (0.5, -0.4)
        self.EYE_IMAGE_POS_L_1 = (-0.5, -0.4) # Used for the two additional fusion images
        self.EYE_IMAGE_POS_R_1 = (0.5, -0.4)

        # Parameters for tracking monitor (norm units)
        self.EYE_SIZE = 0.03
        self.EYE_COLOR_VALID = green
        self.EYE_COLOR_INVALID = red
        self.TRACKING_MONITOR_SIZE = [0.5, 0.5]
        self.TRACKING_MONITOR_POS = [0, 0.4]
        self.TRACKING_MONITOR_COLOR = [0.2, 0.2, 0.2]




# -*- coding: utf-8 -*-
"""
Created on Tue Jun 06 16:20:36 2017

@author: marcus
"""
from psychopy import visual
import numpy as np
import copy
import warnings

HAS_PSYCHOPY = False
try:
    from psychopy.tools.monitorunittools import cm2deg, deg2pix
except:
    pass
else:
    HAS_PSYCHOPY = True


#%%
def tobii2norm(pos):
    ''' Converts from Tobiis coordinate system [0, 1] to PsychoPy's 'norm' (-1, 1).
    Note that the Tobii coordinate system start in the upper left corner
    and the PsychoPy coordinate system in the center.
    y = -1 in PsychoPy coordinates means bottom of screen
    Args:   pos: N x 2 array with positions
    '''

    pos_temp = copy.deepcopy(pos) # To avoid that the called parameter is changed

    # Convert between coordinate system
    pos_temp[:, 0] = 2.0 * (pos_temp[:, 0] - 0.5)
    pos_temp[:, 1] = 2.0 * (pos_temp[:, 1] - 0.5) * -1

    return pos_temp

def norm2tobii(pos):
    ''' Converts from PsychoPy's 'norm' (-1, 1) to Tobiis coordinate system [0, 1].
    Note that the Tobii coordinate system start in the upper left corner
    and the PsychoPy coordinate system in the center.
    y = -1 in PsychoPy coordinates means bottom of screen
    Args:   pos: N x 2 array with positions
    '''
    pos_temp = copy.deepcopy(pos) # To avoid that the called parameter is changed

    # Convert between coordinate system
    pos_temp[:, 0] = pos_temp[:, 0] / 2.0 + 0.5
    pos_temp[:, 1] = pos_temp[:, 1]/ -2.0 + 0.5


    return pos_temp

def tobii2deg(pos, mon):
    ''' Converts Tobiis coordinate system [0, 1 to degrees.
    Note that the Tobii coordinate system start in the upper left corner
    and the PsychoPy coordinate system in the center
    Assumes pixels are square
    Args:   pos: N x 2 array with calibratio position in [0, 1]
            screen_height: height of screen in cm
    '''

    pos_temp = copy.deepcopy(pos) # To avoid that the called parameter is changed

    # Center
    pos_temp[:, 0] = pos_temp[:, 0] - 0.5
    pos_temp[:, 1] = (pos_temp[:, 1] - 0.5) * -1

    # Cenvert to psychopy coordinates (center)
    pos_temp[:, 0] = pos_temp[:, 0] * mon.getWidth()
    pos_temp[:, 1] = pos_temp[:, 1] * mon.getWidth() * (float(mon.getSizePix()[1]) / \
                                              float(mon.getSizePix()[0]))

    # Convert to deg.
    pos_deg = cm2deg(pos_temp, mon, correctFlat=False)
    return pos_deg

def deg2tobii(pos, mon):
    ''' Converts from degrees to Tobiis coordinate system [0, 1].
    Note that the Tobii coordinate system start in the upper left corner
    and the PsychoPy coordinate system in the center
    '''

    # First convert from deg to pixels
    pos[:, 0] = deg2pix(pos[:, 0], mon, correctFlat=False)
    pos[:, 1] = deg2pix(pos[:, 1], mon, correctFlat=False)

    # Then normalize data -1,1
    pos[:, 0] = pos[:, 0] / float(mon.getSizePix()[0])/2
    pos[:, 1] = pos[:, 1] / float(mon.getSizePix()[1])/2

    #.. finally shift to tobii coordinate system
    return norm2tobii(pos)

def tobii2pix(pos, win):
    ''' Converts from  Tobiis coordinate system [0, 1] to pixles.
    Note that the Tobii coordinate system start in the upper left corner
    and screen coordinate in the upper left corner
    Args:   pos: N x 2 array with calibratio position in [0, 1]
            screen_height: height of screen in cm


    '''

    pos_temp = copy.deepcopy(pos)     # Center

    # Center
    pos_temp[:, 0] = pos_temp[:, 0] - 0.5
    pos_temp[:, 1] = (pos_temp[:, 1] - 0.5) * -1

	# Convert to pixels
    pos_temp[:, 0] = pos_temp[:, 0] * win.size[0]
    pos_temp[:, 1] = pos_temp[:, 1] * win.size[1]

    # Convert to deg.
    return pos_temp

def pix2tobii(pos, mon):
    ''' Converts from PsychoPy pixels to Tobiis coordinate system [0, 1].
    Note that the Tobii coordinate system start in the upper left corner
    and the PsychoPy coordinate system in the center
    '''
    # Normalize data -1,1
    pos[:, 0] = pos[:, 0] / (float(mon.getSizePix()[0]) / 2.0)
    pos[:, 1] = pos[:, 1] / (float(mon.getSizePix()[1]) / 2.0)

    #.. finally shift to tobii coordinate system
    return norm2tobii(pos)
#%%
class MyDot2:
    '''
    Generates the best fixation target according to Thaler et al. (2013)
    '''
    def __init__(self, win, outer_diameter=50, inner_diameter=10,
                 outer_color = 'black', inner_color = 'white',units = 'pix'):
        '''
        Class to generate a stimulus dot with
        units are derived from the window
        '''

        # Set propertis of dot
        outer_dot = visual.Circle(win,fillColor = outer_color, radius = outer_diameter/2,
                                  units = units)
        inner_dot = visual.Circle(win,fillColor = outer_color, radius = inner_diameter/2,
                                  units = units)
        line_vertical = visual.Rect(win, width=inner_diameter, height=outer_diameter,
                                    fillColor=inner_color, units = units)
        line_horizontal = visual.Rect(win, width=outer_diameter, height=inner_diameter,
                                    fillColor=inner_color, units = units)


        self.outer_dot = outer_dot
        self.inner_dot = inner_dot
        self.line_vertical = line_vertical
        self.line_horizontal = line_horizontal


    def set_size(self, size):
        ''' Sets the size of the stimulus as scaled by 'size'
        That is, if size == 1, the size is not altered.
        '''
        self.outer_dot.radius = size / 2
        self.line_vertical.size = (self.line_vertical.width, size)
        self.line_horizontal.size = (size, self.line_horizontal.height)

    def set_pos(self, pos):
        '''
        sets position of dot
        pos = [x,y]
        '''
        self.outer_dot.pos = pos
        self.inner_dot.pos = pos
        self.line_vertical.pos = pos
        self.line_horizontal.pos = pos

    def get_pos(self):
        '''
        get position of dot
        '''
        pos = self.outer_dot.pos

        return pos

    def get_size(self):
        '''
        get size of dot
        '''

        return self.outer_dot.radius * 2


    def draw(self):
        '''
        draws the dot
        '''
        self.outer_dot.draw()
        self.line_vertical.draw()
        self.line_horizontal.draw()
        self.inner_dot.draw()


    def invert_color(self):
        '''
        inverts the colors of the dot
        '''
        temp = self.outer_dot.fillColor
        self.outer_dot.fillColor = self.inner_dot.fillColor
        self.inner_dot.fillColor = temp

    def set_color(self, color):
        self.outer_dot.lineColor = 'blue'
        self.outer_dot.fillColor = 'blue'
        self.inner_dot.fillColor = 'blue'
        self.inner_dot.lineColor = 'blue'
        self.line_vertical.lineColor = 'red'
        self.line_horizontal.fillColor = 'red'
        self.line_vertical.fillColor = 'red'
        self.line_horizontal.lineColor = 'red'

#%%
def ellipse(xy = (0, 0), width=1, height=1, angle=0, n_points=50):
    ''' Generates edge points for an ellipse
    Args:
        xy - center of ellipse
        width - width of ellipse
        height - height of ellipse
        angle - angular rotation of ellipse (in radians)
        n_points - number of points to generate

    Return:
        points - n x 2 array with ellipse points
    '''

    xpos,ypos=xy[0], xy[1]
    radm,radn=width,height
    an=angle

    co,si=np.cos(an),np.sin(an)
    the=np.linspace(0,2*np.pi,n_points)
    X=radm*np.cos(the)*co-si*radn*np.sin(the)+xpos
    Y=radm*np.cos(the)*si+co*radn*np.sin(the)+ypos

    points = np.vstack((X, Y)).T

    return points
#%%
class EThead(object):
    """ A class to handle head animation in Titta
    The animated head should reflect a mirror image of the participants
    head.
    """

    def __init__(self, win):
        '''
        Args:
            win - psychopy window handle
        '''

        self.win = win

        # Define colors
        blue = tuple(np.array([37, 97, 163]) / 255.0 * 2 - 1)
        green = tuple(np.array([0, 120, 0]) / 255.0 * 2 - 1)
        red = tuple(np.array([150, 0, 0]) / 255.0 * 2 - 1)
        yellow = tuple(np.array([255, 255, 0]) / 255.0 * 2 - 1)
        yellow_linecolor = tuple(np.array([255, 255, 0]) / 255.0 * 2 - 1)

        # Head parameters
        HEAD_POS_CIRCLE_FIXED_COLOR = blue
        HEAD_POS_CIRCLE_FIXED_RADIUS = 0.20
        self.HEAD_POS_ELLIPSE_MOVING_HEIGHT = 0.20
        self.HEAD_POS_ELLIPSE_MOVING_MIN_HEIGHT = 0.05

        # Eye parameters
        self.EYE_SIZE = 0.03


        # Setup control circles for head position
        self.static_circ = visual.Circle(win, radius = HEAD_POS_CIRCLE_FIXED_RADIUS,
                                         lineColor = HEAD_POS_CIRCLE_FIXED_COLOR,
                                         fillColor = None,
                                         lineWidth=4, units='height')

        # Size=None (If None is specified, the size will be set with values passed to width and height)
        self.moving_ellipse = visual.ShapeStim(win,  lineColor = yellow_linecolor,
                                         lineWidth=4, units='height',
                                         size=None,
                                         fillColor=yellow, opacity=0.5)


        # Ellipses for eyes
        self.eye_l = visual.ShapeStim(win,  lineColor = 'white', fillColor='white',
                                         lineWidth=2, units='height')
        self.eye_r = visual.ShapeStim(win,  lineColor = 'white', fillColor='white',
                                         lineWidth=2, units='height')

        # Ellipses for pupils
        self.pupil_l = visual.ShapeStim(win,  fillColor = 'black',
                                        lineColor = 'black',
                                        units='height')
        self.pupil_r = visual.ShapeStim(win,  fillColor = 'black',
                                        lineColor = 'black',
                                        units='height')

        self.eye_l_closed = visual.Rect(win, fillColor=(1,1,1),
                                        lineColor=(1,1,1), units='height')
        self.eye_r_closed = visual.Rect(win, fillColor=(1,1,1),
                                        lineColor=(1,1,1), units='height')

        self.head_width = 0.25
        self.head_height = 0.25

        # state variables
        self.latest_valid_yaw = 0.
        self.latest_valid_roll = 0.
        self.previous_binocular_sample_valid = True
        self.latest_valid_binocular_avg = np.array([0.5, 0.5, 0.5])
        self.offset = np.array([0., 0., 0.])


    def update(self, sample, sample_user_pos, eye='both'):
        '''
        Args:
            sample - a dict containing information about the sample

                relevant info in sample is

                sample['left_gaze_origin_in_user_coordinate_system']
                sample.right.gaze_origin.in_user_coordinate_system
                sample.left.gaze_origin.in_trackbox_coordinate_system
                sample.right.gaze_origin.in_trackbox_coordinate_system
                sample.left.pupil_diameter
                sample.right.pupil_diameter

            sample_user_pos

                relevant info in sample is

                sample_user_pos.left.user_position
                sample_user_pos.left.user_validity.valid.value
                sample_user_pos.right.user_position
                sample_user_pos.left.user_validity.valid.value

            eye - track, both eyes, left eye, or right eye
                  the non-tracked eye will be indicated by a cross

            latest_binocular_avg
        '''

        self.eye = eye # Which eye(s) should be tracked

        # Indicate that eye is not used by red color
        if 'right' in self.eye:
            self.eye_l_closed.fillColor = (1, -1, -1) # Red

        if 'left' in self.eye:
            self.eye_r_closed.fillColor = (1, -1, -1)

        #%% 1. Compute the average position of the head ellipse
        xyz_pos_eye_l = (sample_user_pos['left_user_position_x'][0],
                         sample_user_pos['left_user_position_y'][0],
                         sample_user_pos['left_user_position_z'][0])

        xyz_pos_eye_r = (sample_user_pos['right_user_position_x'][0],
                         sample_user_pos['right_user_position_y'][0],
                         sample_user_pos['right_user_position_z'][0])

        # Valid data from the eyes?
        self.right_eye_valid = np.sum(np.isnan(xyz_pos_eye_r)) == 0 # boolean
        self.left_eye_valid = np.sum(np.isnan(xyz_pos_eye_l)) == 0

        # Compute the average position of the eyes
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            avg_pos = np.nanmean([xyz_pos_eye_l, xyz_pos_eye_r], axis=0)
        # print('avg pos  {:f} {:f} {:f}'.format(avg_pos[0], avg_pos[1], avg_pos[2]))

        # If one eye is closed, the center of the circle is moved,
        # Try to prevent this by compensating by an offset

        if eye == 'both':
            if self.left_eye_valid and self.right_eye_valid: # if both eyes are open
                self.latest_valid_binocular_avg = avg_pos[:]
                self.previous_binocular_sample_valid = True
                self.offset = np.array([0., 0., 0.])
            elif self.left_eye_valid or self.right_eye_valid:

                if self.previous_binocular_sample_valid:
                    self.offset = self.latest_valid_binocular_avg - avg_pos

                self.previous_binocular_sample_valid = False

        #(0.5, 0.5, 0.5)  means the eye is in the center of the box
        self.moving_ellipse.pos = ((avg_pos[0] - 0.5) * -1 - self.offset[0] ,
                                (avg_pos[1] - 0.5) * -1 - self.offset[1])

        # Compute roll and yaw data from 3D information about the eyes
        # in the headbox (if both eyes are valid)
        if self.left_eye_valid and self.right_eye_valid:
            roll = np.math.tan((xyz_pos_eye_l[1] - xyz_pos_eye_r[1]) / \
                               (xyz_pos_eye_l[0] - xyz_pos_eye_r[0]))
            yaw =  np.math.tan((xyz_pos_eye_l[2] - xyz_pos_eye_r[2]) / \
                               (xyz_pos_eye_l[0] - xyz_pos_eye_r[0])) *-1

            self.latest_valid_roll = roll
            self.latest_valid_yaw = yaw

        else: # Otherwise use latest valid measurement
            roll = self.latest_valid_roll
            yaw = self.latest_valid_yaw
        self.latest_valid_roll_deg = roll * 180 / np.pi * -1

#        print('test', latest_valid_binocular_avg, roll, yaw)

        # Compute the ellipse height and width
        # The width should be zero if yaw = pi/2 rad (90 deg)
        # The width should be equal to the height if yaw = 0
        ellipse_height = (avg_pos[2] - 0.5)*-1 * 0.2 + self.HEAD_POS_ELLIPSE_MOVING_HEIGHT

        # Control min size of head ellipse
        if ellipse_height < self.HEAD_POS_ELLIPSE_MOVING_MIN_HEIGHT:
            ellipse_height = self.HEAD_POS_ELLIPSE_MOVING_MIN_HEIGHT

        self.head_width = ellipse_height - \
                          np.abs(yaw) / np.pi * (ellipse_height)

#        print(self.moving_ellipse.pos, self.moving_ellipse.height,
#              self.head_width, roll, yaw)

        # Get head ellipse points to draw
        ellipse_points_head = ellipse(xy = (0, 0),
                                      width= self.head_width,
                                      height=ellipse_height,
                                      angle=roll)

        # update position and shape of head ellipse
        self.moving_ellipse.vertices = ellipse_points_head

        #%% Compute the position and size of the eyes (roll)
        eye_head_distance = self.head_width / 2
        self.eye_l.pos = (self.moving_ellipse.pos[0] - np.cos(roll) * eye_head_distance,
                          self.moving_ellipse.pos[1] - np.sin(roll) * eye_head_distance)
        self.eye_l.vertices = ellipse_points_head / (5.0 - yaw)

        self.eye_r.pos = (self.moving_ellipse.pos[0] + np.cos(roll) * eye_head_distance,
                          self.moving_ellipse.pos[1] + np.sin(roll) * eye_head_distance)
        self.eye_r.vertices = ellipse_points_head / (5.0 + yaw)

        #%% Compute the position and size of the pupils
#        print(self.eye_l.pos)
        self.pupil_l.pos = self.eye_l.pos
        self.pupil_l.vertices = self.eye_l.vertices * (sample['left_pupil_diameter'][0] - 1)*2 / 16

        self.pupil_r.pos = self.eye_r.pos
        self.pupil_r.vertices = self.eye_r.vertices * (sample['right_pupil_diameter'][0] - 1)*2 / 16

#        print(self.eye_l.pos, self.eye_r.pos)

        return self.latest_valid_binocular_avg, self.previous_binocular_sample_valid, self.latest_valid_roll, self.latest_valid_yaw, self.offset

    def draw(self):
        ''' Draw all requested features
        '''

        # draw reference
        self.static_circ.draw()

        # early out if nothing else to draw
        if not (self.right_eye_valid or self.left_eye_valid):
            return

        # Draw head, eyes, and pupils
        self.moving_ellipse.draw()

#        print(self.eye, self.moving_ellipse.pos, self.eye_r.pos, self.pupil_r.vertices)
        if 'both' in self.eye or 'right' in self.eye:
            if self.right_eye_valid:
                self.eye_r.draw()
                self.pupil_r.draw()
            else:
                self.eye_r_closed.pos = self.eye_r.pos
                self.eye_r_closed.ori = self.latest_valid_roll_deg
                self.eye_r_closed.size = (self.head_width / 2.0, self.head_width / 8.0)
                self.eye_r_closed.draw()

            if 'right' in self.eye:
                # Indicate that the left eye is not used
                self.eye_l_closed.pos = self.eye_l.pos
                self.eye_l_closed.ori = self.latest_valid_roll_deg
                self.eye_l_closed.size = (self.head_width / 2.0, self.head_width / 8.0)
                self.eye_l_closed.draw()


        if 'both' in self.eye or 'left' in self.eye:
            if self.left_eye_valid:
                self.eye_l.draw()
                self.pupil_l.draw()
            else:
                self.eye_l_closed.pos = self.eye_l.pos
                self.eye_l_closed.ori = self.latest_valid_roll_deg
                self.eye_l_closed.size = (self.head_width / 2.0, self.head_width / 8.0)
                self.eye_l_closed.draw()

            if 'left' in self.eye:
                # Indicate that the right eye is not used
                self.eye_r_closed.pos = self.eye_r.pos
                self.eye_r_closed.ori = self.latest_valid_roll_deg
                self.eye_r_closed.size = (self.head_width / 2.0, self.head_width / 8.0)
                self.eye_r_closed.draw()


#%%
class AnimatedCalibrationDisplay(object):
    """ A class for drawing animated targets"""
    def __init__(self, win, target, function_name):
        ''' The function 'function_name' does the actual drawing
        '''
        self.win = win
        self.function_name = function_name
        self.target = target # psychopy.visual object (should be in 'pix' units)
        self.target_size = target.get_size()
        self.screen_refresh_rate = float(win.getActualFrameRate() or 60)

    def animate_target(self, point_number, position, tick):
        ''' Calls the target drawing function 'func'
        '''

        eval(''.join(['self.',self.function_name,'(',str(point_number),',',
                                        '(',str(position[0]),',',
                                            str(position[1]), ')' ,',', str(tick),')']))


    def animate_point(self, point_number, position, tick):
        ''' Animates calibration point with a certain point_number and position
        tick increases by 1 for each call
        Args:
            position - (x, y)
        '''
        target_size = np.abs(self.target_size - self.target_size * np.sin(3 * tick / self.screen_refresh_rate + 4*np.pi/2)) + self.target_size / 5
        self.target.set_size(target_size)
        self.target.set_pos(position)
        self.target.draw()


    def move_point(self, old_position, new_position, tick):
        ''' Animates movement between two positions
        '''
        move_completed = False

        # The target should have a fixed size when moving
        self.target.set_size(2 * self.target_size)

        # How many ticks should the movement be (one screen unit in one second)?
        n_steps = int(self.screen_refresh_rate / 2)
        step_pos_x = np.linspace(old_position[0], new_position[0], n_steps)
        step_pos_y = np.linspace(old_position[1], new_position[1], n_steps)


        if tick >= len(step_pos_x):
            move_completed = True
        else:
            self.target.set_pos((step_pos_x[tick], step_pos_y[tick]))

        self.target.draw()

        return move_completed


#%%
def unit_vector(vector):
    """ Returns the unit vector of the vector.  """
    return vector / np.linalg.norm(vector)

#%%
def angle_between(v1, v2):
    """ Returns the angle in radians between vectors 'v1' and 'v2'::

            >>> angle_between((1, 0, 0), (0, 1, 0))
            1.5707963267948966
            >>> angle_between((1, 0, 0), (1, 0, 0))
            0.0
            >>> angle_between((1, 0, 0), (-1, 0, 0))
            3.141592653589793
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))

#%%
def rms(x):
    rms = np.sqrt(np.mean(np.square(np.diff(x))))
    return rms

#%%
def sd(x):
    return np.std(x)

#%%




# -*- coding: utf-8 -*-
"""
Created on Thu Jun 01 14:11:57 2017

@author: Marcus
"""

from psychopy import visual, core, event
import time
import pandas as pd
import h5py
import os
from pathlib import Path
from titta import helpers_tobii as helpers
from threading import Thread
import numpy as np
import sys

sample_list = ['device_time_stamp', 'system_time_stamp',
       'left_gaze_on_display_area_x',
       'left_gaze_on_display_area_y',
       'left_gaze_point_in_user_coordinates_x',
       'left_gaze_point_in_user_coordinates_y',
       'left_gaze_point_in_user_coordinates_z', 'left_gaze_point_valid',
       'left_gaze_point_available', 'left_pupil_diameter', 'left_pupil_valid',
       'left_pupil_available', 'left_gaze_origin_in_user_coordinates_x',
       'left_gaze_origin_in_user_coordinates_y',
       'left_gaze_origin_in_user_coordinates_z',
       'left_gaze_origin_in_track_box_coordinates_x',
       'left_gaze_origin_in_track_box_coordinates_y',
       'left_gaze_origin_in_track_box_coordinates_z', 'left_gaze_origin_valid',
       'left_gaze_origin_available', 'left_eye_openness_diameter',
       'left_eye_openness_valid', 'left_eye_openness_available',
       'right_gaze_on_display_area_x',
       'right_gaze_on_display_area_y',
       'right_gaze_point_in_user_coordinates_x',
       'right_gaze_point_in_user_coordinates_y',
       'right_gaze_point_in_user_coordinates_z', 'right_gaze_point_valid',
       'right_gaze_point_available', 'right_pupil_diameter',
       'right_pupil_valid', 'right_pupil_available',
       'right_gaze_origin_in_user_coordinates_x',
       'right_gaze_origin_in_user_coordinates_y',
       'right_gaze_origin_in_user_coordinates_z',
       'right_gaze_origin_in_track_box_coordinates_x',
       'right_gaze_origin_in_track_box_coordinates_y',
       'right_gaze_origin_in_track_box_coordinates_z',
       'right_gaze_origin_valid', 'right_gaze_origin_available',
       'right_eye_openness_diameter', 'right_eye_openness_valid',
       'right_eye_openness_available']

class Buffer(Thread):
    """
    Create that simulates a data buffer
    """
    def __init__(self, win):

        self.win = win

        # Mouse object
        self.mouse = event.Mouse(win)
        self.mouse.setPos([0,0])
        self.mouse.setVisible(0)

        self.__stop = True
        self.sample = {}

        # Initiate an empty sample dict
        for l in sample_list:
            self.sample[l] = [0]

        self._start_sample_buffer()

    #%%
    def peek_N(self, stream, N):
        ''' Consume all samples and empty buffer'''
        return self.sample

    #%%
    def consume_N(self, stream, N):
        ''' Get samples in buffer while emptying the buffer '''

        temp = self.sample
        self.sample = {}
        return temp

    #%%
    def _start_sample_buffer(self, sample_buffer_length=sys.maxsize):

        Thread.__init__(self)

        # Initialize the ring buffer
        self.__stop = False
        self.start()


    #%%
    def run(self):
        # Called by the e.g., et.start()
        # Continously read data into a ringbuffer
        while True:
            if self.__stop:
                break

            # Add sample to dict
            self._add_sample_to_buffer()
            time.sleep(0.01)

    #%%
    def _add_sample_to_buffer(self):
        ''' Simulates gaze position with mouse
        '''
        x, y = self.mouse.getPos()

        # Convert mouse position in current psychopy units to Tobii's
        # norm coordinate system (currently supports 'norm' and 'deg')

        if self.win.units == 'norm':
            xy = helpers.norm2tobii(np.array([x, y],ndmin=2))
        elif self.win.units == 'deg':
            xy = helpers.deg2tobii(np.array([x, y],ndmin=2), self.win.monitor)
        elif self.win.units == 'pix':
            xy = helpers.pix2tobii(np.array([x, y],ndmin=2), self.win.monitor)
        else:
            raise IOError ('Invalid unit of PsychoPy screen: Titta in dummy mode currently \
                           supports "norm", "pix", and "deg".')
        for key in self.sample.keys():
            if '_x' in key:
                self.sample[key].append(xy[0, 0])
            elif '_y':
                self.sample[key].append(xy[0, 1])
            else:
                self.sample[key].append(np.random.rand())

    #%%
    def _stop_sample_buffer(self):
        self.__stop = True
##########################
##########################
##########################
##########################

# %%
class Connect(object):
    """
    Create a class that simplifies life for people wanting to use the SDK
    """
    def __init__(self):

        # clock
        self.clock = core.Clock()

        # Dict to store samples

    def init(self):
        ''' Connect to eye tracker
        and apply settings
        '''
        print('init')


    #%% Init calibration
    def calibrate(self, win, eye='both', calibration_number='first'):
        ''' Master function for setup and calibration
        '''
        self.buffer = Buffer(win)

        # Window and instruction text for calibration
        instruction_text = visual.TextStim(win,text='',wrapWidth = 1,
                                           height = 0.05, units='norm')

        self.instruction_text = instruction_text

        self.win = win

        self.instruction_text.text = 'Calibration Dummy mode'
        self.instruction_text.draw()
        self.win.flip()
        core.wait(2)

    #%%
    def get_system_time_stamp(self):
        ''' Get system time stamp
        '''

        print('time_stamp')

    #%%
    def start_recording(self,   gaze=False,
                                time_sync=False,
                                eye_image=False,
                                notifications=False,
                                external_signal=False,
                                positioning=False):
        print('start_recording')



    #%%
    def send_message(self, msg, ts=None):
        print(msg)


    #%%
    def stop_recording(self,    gaze=False,
                                time_sync=False,
                                eye_image=False,
                                notifications=False,
                                external_signal=False,
                                positioning=False):
        print('stop_recording')

    #%%
    def set_dummy_mode(self):
        print('dummy mode')

    #%%
    def calibration_history(self):
        print('calibration_history')
        return None
    #%%
    def system_info(self):
        ''' Returns information about system in dict
        '''

        info = {}
        info['serial_number']  = 'dummy'
        info['address']  = 'dummy'
        info['model']  = 'dummy'
        info['name']  = 'dummy'
        info['firmware_version'] = 'dummy'
        info['runtime_version'] = 'dummy'
        info['tracking_mode']  = 'dummy'
        info['sampling_frequency']  = 'dummy'
        info['track_box']  = 'dummy'
        info['display_area']  = 'dummy'
        info['python_version'] = '.'.join([str(sys.version_info[0]),
                                           str(sys.version_info[1]),
                                           str(sys.version_info[2])])
        info['psychopy_version'] = 'dummy'
        info['TittaPy_version'] = 'dummy'
        info['titta_version'] = 'dummy'

    #%%
    def set_sample_rate(self, Fs):
        '''Sets the sample rate
        '''
        print('set sample rate')
    #%%
    def get_sample_rate(self):
        '''Gets the sample rate
        '''
        print('get sample rate')
        return 0

    #%%
    def save_data(self, *argv, filename=None, append_version=True):
        ''' Saves the data to HDF5 container
        If you want to read the data, see the 'resources' folder
        '''

        print('save data')







# -*- coding: utf-8 -*-
"""
Created on Thu Dec  6 14:37:06 2018

@author: Marcus
ToDo: upload information in chunks of 64 KB
"""

# import relevant libs
from websocket import create_connection
import json
import os
import time
import numpy as np
import threading

    
status_codes = {0: 'Operation successful',
                100: 'Bad request',
                101: 'Invalid parameter',
                102: 'Operation was unsuccessful',
                103: 'Operation cannot be executed in current state',
                104: 'Access to the service is forbidden',
                105: 'Authorization during connection to a service has not been provided',
                201: 'Recording finalization failed'}

class TalkToProLab(threading.Thread):
    """ A class for talking to Tobii Pro Lab. 
    Tested with Python 2.7 and 3.6
    """     
    
    #%%
    def __init__(self, project_name=None, dummy_mode=False):
        
        if dummy_mode:
            from titta import TalkToProLab_dummy
            self.__class__ = TalkToProLab_dummy.TalkToProLab_dummy
            self.__class__.__init__(self) 
        else:
            # Connect to servers
            self.clock_address = create_connection('ws://localhost:8080/clock?client_id=RemoteClient')
            self.external_presenter_address = create_connection('ws://localhost:8080/record/externalpresenter?client_id=RemoteClient')
            self.project_address = create_connection('ws://localhost:8080/project?client_id=RemoteClient')
            
            # Make sure the desired project is opened (if project name is given)
            if project_name:
                assert project_name == self.get_project_info()['project_name'], \
                "Wrong project opened in Lab. Should be {}".format(project_name)
    
                
            # Start new thread that takes care of keeping the connection alive
            threading.Thread.__init__(self)
            self.__stop = False
            self.start()
 
    #%%s
    def run(self):
        ''' Ping/pong the server every 15 s to keep the connection alive.
        Starts the thread that keeps the connection alive.
        Otherwise, if there is not communication between the server and client,
        the connection dies after 30 sec.
        '''
        
        while True:
            self.clock_address.ping()
            self.clock_address.pong("pong")
            
            self.external_presenter_address.ping()
            self.external_presenter_address.pong("pong")
            
            self.project_address.ping()
            self.project_address.pong("pong")
            
            time.sleep(15)
            
            if self.__stop == True:
                break
            
    #%% 
    def send_message(self, address, msg_dict, to_json=True):
        ''' Send message to server and return response
        '''
    
        # Convert message dictionary to json
        if to_json:
            msg_dict = json.dumps(msg_dict)
            
        # Send data and wait for response
        address.send(msg_dict)
        response = address.recv()
        
        # Return dict
        if to_json:
            return json.loads(response)
        else:
            return response

    #%%
    def get_api_version(self):
        ''' Get version of api 
         Available for all Services. 
        request:
        {"operation": "GetApiVersion"}
        
        response:
        {"operation": "GetApiVersion",
        "status_code": 0,
        "version": "1.0"}        
        '''
        response = self.send_message(self.clock_address, 
                                     {"operation": "GetApiVersion"})
        
        
        assert response['status_code'] == 0, response
        return response        
        
    #%%          
    def get_time_stamp(self):
        ''' Clock API
        request:
        {"operation": "GetTimestamp"}
        
        response:
        {"operation": "GetTimestamp",
        "status_code": 0,
        "timestamp": "128606207528320000"}
        
        Note: timestamp in microseconds
        '''
        
        response = self.send_message(self.clock_address,
                                     {"operation": "GetTimestamp"})
                    
        assert response['status_code'] == 0, response
        
        return response
    #%%
    def add_participant(self, participant_name):
        ''' Project data API
        request:
        {"operation": "AddParticipant",
        "participant_name": "John Doe"}
        
        response:
        {"operation": "AddParticipant",
        "status_code": 0,
        "participant_id": "2A86C895-F8B6-4786-8C4B-CB889C3449F1"}
        '''   
        
        # Make sure that the participant id does not already exist in Lab
        assert not self.find_participant(participant_name), "Participant {} already exists in Lab".format(participant_name)
        
        response = self.send_message(self.project_address,
                                     {"operation": "AddParticipant",
                                      "participant_name": participant_name})
                    
        assert response['status_code'] == 0, response
        
        return response        
        
    #%%
    def get_project_info(self):
        ''' Project data API
        request:
        { "operation": "GetProjectInfo"}
        
        response:
        {"operation": "GetProjectInfo",
        "status_code": 0,
        "project_id": "8928673C-B1D5-4941-A592-39597742B1E5",
        "project_name": "SocialInteraction test"}
        '''    
        
        response = self.send_message(self.project_address,
                                     { "operation": "GetProjectInfo"})
                    
        assert response['status_code'] == 0, response
        
        return response         
    
    #%%
    def find_participant(self, participant_name):
        ''' Finds whether participant_name already has been uploaded to lab
        
        Args:
            participant_name - string with participant name (e.g., P04)
        
        Returns:
            exists - boolean 
        '''
        
        exists = False
        uploaded_participants = self.list_participants()['participant_list']
        for m in uploaded_participants:
            if m['participant_name'] == participant_name: 
                exists = True
        
        return exists
    
    #%% 
    def list_participants(self):
        ''' Project data API
        Returns information about the participants of the currently opened project in Pro Lab.
        
        request:
        {
        "operation": "ListParticipants"
        } 
        response:
        {
        "operation": "ListParticipants",
        "status_code": 0,
        "participant_list": [{
        "participant_name": "John Doe",
        "participant_id": "2A86C895-F8B6-4786-8C4B-CB889C3449F1"
        }]
        }        
        '''
        
        response = self.send_message(self.project_address,
                                     { "operation": "ListParticipants"})
                    
        assert response['status_code'] == 0, response
        
        return response            
    
    #%% 
    def list_media(self):
        ''' Project data API
        Returns information about the uploaded media of the currently opened project in Pro Lab.
        
        request:
        {
        "operation": "ListMedia"
        }        
        
        response:
        {
        "operation": "ListMedia",
        "status_code": 0,
        "media_list": [{
            "media_name": "CuteCats",
            "media_id": "b2548e0e-4056-4d36-a4f5-86b3194866f7",
            "mime_type": "image/jpeg",
            "media_size": 3536263,
            "width": 2512,
            "height": 1884,
            "duration": 0,
        },
        {
            "media_name": "CuteCatsRunningAround",
            "media_id": "21BA2272-473D-478F-9E23-A1FA8ABFF65D",
            "md5_checksum": "0dbc22fddaff860ac1bf0d092c909d4e",
            "mime_type": "video/mp4",
            "media_size": 3536263,
            "width": 640,
            "height": 480,
            "duration": 5.250
        }]
        }   
        '''
        
        response = self.send_message(self.project_address,
                                     { "operation": "ListMedia"})
                    
        assert response['status_code'] == 0, response
        
        return response      
       
        
    #%%
    def find_media(self, media_name):
        ''' Finds whether media_name already has been uploaded to lab
        
        Args:
            media_name - string with media name (e.g., image.png)
        
        Returns:
            exists - boolean 
        '''
        
        exists = False
        uploaded_media = self.list_media()['media_list']
        for m in uploaded_media:
            if m['media_name'] == media_name.split('.')[0]:
                exists = True
        
        return exists
        
    #%%
    def upload_media(self, media_name, media_type):
        ''' 
        
        Prepares Server to receive media content (image or video) as binary data.
        The response will be sent once immediately after request and once at the end of the upload
        operation, regardless the reason of upload operation completion. After a success response to an
        initial request, chunked binary data is expected to be sent.
        The whole operation of receiving media could be ended due to the following reasons:
            
        1. The Server received an amount of binary data equal to media_size parameter.
        The operation has been aborted by  
        2. UploadMediaAbort request.
        3. The time between the Server received two consecutive chunks exceeded timeout.  
            -> status 103
            
        Args:
            media_name - name of media.ext (e.g., image.png)
            media_type - 'image' or 'video'
            
        
        
        -----        
        Project data API
        request:
        {"operation": "UploadMedia",
        "mime_type": "video/mp4",
        "media_name": "SocialInteractionDemoVideo.mp4",
        "media_size": "37201059"}
        
        response:
        {"operation": "UploadMedia",
        "status_code": 0}
        '''   
        
        # File extension
        file_ext = media_name.split('.')[1]
        
        # Check that media format is supported
        assert(len([i for i in ['bmp','jpeg', 'png','gif','mp4', 'x-msvideo'] \
                    if file_ext == i]) > 0), "File type not supported"
        
        # Read image as binary, each element in 'f' represents a byte
        with open(media_name, "rb") as imageFile:
          f = imageFile.read()
          
        # Prepare Server to receive media content (image or video) as binary data          
        media_size = len(f)
        response = self.send_message(self.project_address,
                                     {"operation": "UploadMedia",
                                      "mime_type": '/'.join([media_type, 
                                                             file_ext]),
                                      "media_name": media_name,
                                      "media_size": str(media_size)})
        assert response['status_code'] == 0, response
          
        # Send data to server (Should be sent in 64KB chunks)
        c_idx = np.arange(0, media_size, 64000)
        for i in np.arange(len(c_idx))[:-1]:
            self.project_address.send_binary(f[c_idx[i] : (c_idx[i + 1])])    
            
        # Send last bits
        self.project_address.send_binary(f[c_idx[-1] : media_size])  
            
      
        # self.project_address.send_binary(f)        
                      
        # Wait for verification that all bytes were received by lab
        response =  json.loads(self.project_address.recv())
        '''
        response:
        {
        "operation": "UploadMedia",
        "status_code": 0,
        "md5_checksum": "7815696ecbf1c96e6894b779456d330e",
        "media_id": "2A86C895-F8B6-4786-9C4B-CB889C3449F0"
        }     
        '''
          
        
        assert response['status_code'] == 0, response
        
        return response
          
    #%% TODO: remove?
    def upload_media_abort(self):
        ''' Project data API
        request:
        {"operation": "UploadMediaAbort"}
        
        response:
        {"operation": "UploadMediaAbort",
        "status_code": 0}
        '''    
        
        response = self.send_message(self.project_address,
                                      {"operation": "UploadMediaAbort"})  
        
        assert response['status_code'] == 0, response
        
        return response
 
    #%%
    def add_aois_to_image(self, media_id, aoi_name, aoi_color, 
                  key_frame_vertices, key_frame_active = True, 
                  key_frame_seconds = 0, tag_name = '',
                  group_name='', merge_mode='replace_aois'):
        
        
        ''' Project data API
        request:
    
        {
        "operation": "AddAois",
        "media_id": "b3418558-b245-4185-9bdc-725fb6a23a88",
        "aois":  {'name:' : 'test',
                  'color' : 'AACC33',
                  'key_frames' : {'is_active' : true,
                                  'seconds' : 0,
                                  'vertices' : {{100, 100},
                                                {100, 200},
                                                {200, 200},
                                                {200, 100}}},
                  'tags': {'tag_name' : 'vehicle',
                          'group_name' : 'Brand'}},
        "merge_mode": "ReplaceAois"
        }  
                
        response:
        {
        "operation": "sendAois",
        "imported_aoi_count": 1
        }            
        ''' 
        
        s = ''
        for i, k in enumerate(key_frame_vertices):
            
            if i == (len(key_frame_vertices) - 1):
                s = s + '{"x": %s, "y": %s}'
            else:
                s = s + '{"x": %s, "y": %s},'
                
                
        a = (media_id, aoi_name, aoi_color, str(key_frame_active).lower(),
                                key_frame_seconds) + \
                                tuple([x for b in key_frame_vertices for x in b]) + \
                                (group_name, tag_name, merge_mode)
                                
        
               
                                  
        operation = '''{
                        "operation": "AddAois",
                        "media_id": "%s",
                        "aois": [{
                                "name": "%s",
                                "color": "%s",
                                "key_frames": [{
                                        "is_active": %s,
                                        "seconds": %s,
                                        "vertices": [''' + \
                                        s + \
                        '''
                                                ]
                                        }],
                                "tags": [
                                         {
                                            "group_name": "%s",
                                            "tag_name": "%s"
                                            }
                                        ]
                                }],
                                        "merge_mode": "%s"
                        }''' 
            
                                 
                      
        operation = str(operation)                              
        operation = operation.replace('\t', '')
        operation = operation.replace('\n', '')
        operation = operation.replace(' ', '')
        
        operation = operation % a       
            

        response = json.loads(self.send_message(self.project_address,
                                      operation, to_json=False))
        assert response['status_code'] == 0, response
        
        return response
    
   #%%
    def add_aois_to_video(self, media_id, aoi_name, aoi_color, 
                  key_frame_vertices, key_frame_active = True, 
                  key_frame_seconds = 0, tag_name = '',
                  group_name='', merge_mode='replace_aois'):
        ''' 
        
        Placeholder for adding AOIs to video. RTFM!
        
        Project data API
        request:
    
        {
        "operation": "AddAois",
        "media_id": "b3418558-b245-4185-9bdc-725fb6a23a88",
        "aois":  {'name:' : 'test',
                  'color' : 'AACC33',
                  'key_frames' : {'is_active' : true,
                                  'seconds' : 0,
                                  'vertices' : {{100, 100},
                                                {100, 200},
                                                {200, 200},
                                                {200, 100}}},
                  'tags': {'tag_name' : 'vehicle',
                          'group_name' : 'Brand'}},
        "merge_mode": "ReplaceAois"
        }  
                
        response:
        {
        "operation": "sendAois",
        "imported_aoi_count": 1
        }            
        '''     
    #%%
    def get_state(self):
        ''' External presenter API.
        Returns the current state of External Presenter workflow.
        
        request:
        {
        "operation": "GetState"
        }
        response:
        {
        "operation": "GetState",
        "status_code": 0,
        "state": "unmet"
        }        
        '''
		
        response = self.send_message(self.external_presenter_address,
                                      {"operation": "GetState"})  
        
        assert response['status_code'] == 0, response

        return response

    #%%           
    def start_recording(self, recording_name, participant_id, screen_width,
                        screen_height, screen_latency=10000):
        ''' External presenter API     
        request:
        {"operation": "StartRecording",
        "recording_name": "Waterfall",
        "participant_id": "3E6C3751-389A-47F0-861D-7A09D744FC23",
        "screen_latency": 10000,
        "screen_width": 1024,
        "screen_height": 768}
        
        response:
        {"operation": "StartRecording",
        "status_code": 0,
        "recording_id": "4B7F3751-389A-47D0-861D-7A09D744FC12"}
        
        Note: Participant identifier (returned in
                                      AddParticipant operation
                                      response)
        '''
        
        response = self.send_message(self.external_presenter_address,
                                     {"operation": "StartRecording",
                                      "recording_name": recording_name,
                                      "participant_id": participant_id,
                                      "screen_latency": screen_latency,
                                      "screen_width": screen_width,
                                      "screen_height": screen_height})
        
        assert response['status_code'] == 0, response
        
        return response
    #%%           
    def stop_recording(self):
        ''' External presenter API     
        request:
        {"operation": "StopRecording"}
        
        response:
        {"operation": "StopRecording",
        "status_code": 0}
        '''   
        
        response = self.send_message(self.external_presenter_address,
                                      {"operation": "StopRecording"})
        
        assert response['status_code'] == 0, response
        
    #%%           
    def send_stimulus_event(self, recording_id, start_timestamp, 
                            media_id, media_position=None, background=None,
                            end_timestamp=None):
                            
        ''' External presenter API.
        Timestamps should be in microseconds
        
        request:
        {"operation": "SendStimulusEvent",
        "recording_id": "4B7F3751-389A-47D0-861D-7A09D744FC12",
        "start_timestamp": "123123123123213123",
        "end_timestamp": "321321321321321321",
        "media_id": "2A86C895-F8B6-4786-9C4B-CB889C3449F0",
        "media_position":  {"left": 10,
                            "top": 20,
                            "right": 850,
                            "bottom": 1100},
        "background": "AA91C0"}
        
        response:
        {"operation": "SendStimulusEvent",
        "status_code": 0}
        '''       
        
        request =  {"operation": "SendStimulusEvent",
            "recording_id": recording_id,
            "start_timestamp": start_timestamp,
            "media_id": media_id}
        
        if end_timestamp:
            request =  {"operation": "SendStimulusEvent",
                "recording_id": recording_id,
                "start_timestamp": start_timestamp,
                'end_timestamp': end_timestamp,
                "media_id": media_id}       
            
        if  media_position:
            request['media_position'] = media_position
            
        if background:
            request['background'] = background
            
        response = self.send_message(self.external_presenter_address,
                                     request)
        
        assert response['status_code'] == 0, response
    #%%           
    def send_custom_event(self):
        ''' External presenter API     
        request:
        {"operation": "SendCustomEvent",
        "recording_id": "4B7F3751-389A-47D0-861D-7A09D744FC12",
        "timestamp": "123123123123213123",
        "event_type": "ERROR_SEM_OWNER_DIED",
        "value": "105 (0x69)"}
        
        response:
        {"operation": "SendCustomEvent",
        "status_code": 0}
        '''          
    #%%           
    def finalize_recording(self, recording_id):
        ''' Finalizes the recording and makes it ready for analysis in the 
        application.
        
        If validation of data was sent, Server responds with status code 201. 
        Once the root cause of a validation failure has been fixed, 
        the command can be executed again.        
        
        External presenter API.  
        request:
        {"operation": "FinalizeRecording",
        "recording_id": "4B7F3751-389A-47D0-861D-7A09D744FC12"}
        
        response:
        {"operation": "FinalizeRecording",
        "status_code": 0}
        
        Note: recording_id is returned from start_recording
        '''             
        
        response = self.send_message(self.external_presenter_address,
                                     {"operation": "FinalizeRecording",
                                      "recording_id": recording_id})  
        assert response['status_code'] == 0, response
        
        # Stop ping ponging to keep the connection alive
        self.__stop = True
        
    #%% 
    def disconnect(self):
        ''' Closes the websocket connection
        '''
        self.clock_address.close()
        self.external_presenter_address.close()
        self.project_address.close()
        
        
        


# -*- coding: utf-8 -*-
"""
Created on Thu Dec  6 14:37:06 2018

@author: Marcus
ToDo: upload information in chunks of 64 KB
"""

status_codes = {0: 'Operation successful',
                100: 'Bad request',
                101: 'Invalid parameter',
                102: 'Operation was unsuccessful',
                103: 'Operation cannot be executed in current state',
                104: 'Access to the service is forbidden',
                105: 'Authorization during connection to a service has not been provided',
                201: 'Recording finalization failed'}

class TalkToProLab_dummy(object):
    """ A dummy class for talking to Tobii Pro Lab. 
    """     
    
    #%%
    def __init__(self):
        
        print('Dummy class init')

    #%%
    def get_api_version(self):
        ''' Get version of api 
         Available for all Services. 
        request:
        {"operation": "GetApiVersion"}
        
        response:
        {"operation": "GetApiVersion",
        "status_code": 0,
        "version": "1.0"}        
        '''
        print('get_api_version')

        return {'operation': 'GetApiVersion', 'status_code': 0, 'version': 'dummy'}        
        
    #%%          
    def get_time_stamp(self):
        ''' Clock API
        request:
        {"operation": "GetTimestamp"}
        
        response:
        {"operation": "GetTimestamp",
        "status_code": 0,
        "timestamp": "128606207528320000"}
        
        Note: timestamp in microseconds
        '''
        
        print('get_time_stamp')

        return {"operation": "GetTimestamp", "status_code": 0, "timestamp": "10000"}  
    #%%
    def add_participant(self, participant_name):
        ''' Project data API
        request:
        {"operation": "AddParticipant",
        "participant_name": "John Doe"}
        
        response:
        {"operation": "AddParticipant",
        "status_code": 0,
        "participant_id": "2A86C895-F8B6-4786-8C4B-CB889C3449F1"}
        '''   
        
        print('add_participant')

        return {"operation": "AddParticipant", "status_code": 0, 
                "participant_id": "dummy_participant_id"}    
        
    #%%
    def get_project_info(self):
        ''' Project data API
        request:
        { "operation": "GetProjectInfo"}
        
        response:
        {"operation": "GetProjectInfo",
        "status_code": 0,
        "project_id": "8928673C-B1D5-4941-A592-39597742B1E5",
        "project_name": "SocialInteraction test"}
        '''    
        print('get_project_info')

        return {"operation": "GetProjectInfo",
                "status_code": 0,
                "project_id": "dummy_project_id",
                "project_name": "dummy_name"}       
    
    #%%
    def find_participant(self, participant_name):
        ''' Finds whether participant_name already has been uploaded to lab
        
        Args:
            participant_name - string with participant name (e.g., P04)
        
        Returns:
            exists - boolean 
        '''
        
        print('find_participant')
        
        return False
    
    #%% 
    def list_participants(self):
        ''' Project data API
        Returns information about the participants of the currently opened project in Pro Lab.
        
        request:
        {
        "operation": "ListParticipants"
        } 
        response:
        {
        "operation": "ListParticipants",
        "status_code": 0,
        "participant_list": [{
        "participant_name": "John Doe",
        "participant_id": "2A86C895-F8B6-4786-8C4B-CB889C3449F1"
        }]
        }        
        '''
        
        print('find_participant')
        
        return {
                "operation": "ListParticipants",
                "status_code": 0,
                "participant_list": [{
                        "participant_name": "dummy_participant",
                        "participant_id": "dummy_participant_id"
                        }]
                }          
    
    #%% 
    def list_media(self):
        ''' Project data API
        Returns information about the uploaded media of the currently opened project in Pro Lab.
        
        request:
        {
        "operation": "ListMedia"
        }        
        
        response:
        {
        "operation": "ListMedia",
        "status_code": 0,
        "media_list": [{
            "media_name": "CuteCats",
            "media_id": "b2548e0e-4056-4d36-a4f5-86b3194866f7",
            "mime_type": "image/jpeg",
            "media_size": 3536263,
            "width": 2512,
            "height": 1884,
            "duration": 0,
        },
        {
            "media_name": "CuteCatsRunningAround",
            "media_id": "21BA2272-473D-478F-9E23-A1FA8ABFF65D",
            "md5_checksum": "0dbc22fddaff860ac1bf0d092c909d4e",
            "mime_type": "video/mp4",
            "media_size": 3536263,
            "width": 640,
            "height": 480,
            "duration": 5.250
        }]
        }   
        '''
        
        print('list_media')
        
        return {
                "operation": "ListMedia",
                "status_code": 0,
                "media_list": [{
                    "media_name": "dummy_media_name_1",
                    "media_id": "dummy_media_id_1",
                    "mime_type": "image/jpeg",
                    "media_size": 3536263,
                    "width": 2512,
                    "height": 1884,
                    "duration": 0,
                                }]
                }     
       
        
    #%%
    def find_media(self, media_name):
        ''' Finds whether media_name already has been uploaded to lab
        
        Args:
            media_name - string with media name (e.g., image.png)
        
        Returns:
            exists - boolean 
        '''
        
        print('find_media')
        
        return False
        
    #%%
    def upload_media(self, media_name, media_type):
        ''' 
        
        Prepares Server to receive media content (image or video) as binary data.
        The response will be sent once immediately after request and once at the end of the upload
        operation, regardless the reason of upload operation completion. After a success response to an
        initial request, chunked binary data is expected to be sent.
        The whole operation of receiving media could be ended due to the following reasons:
            
        1. The Server received an amount of binary data equal to media_size parameter.
        The operation has been aborted by  
        2. UploadMediaAbort request.
        3. The time between the Server received two consecutive chunks exceeded timeout.  
            -> status 103
            
        Args:
            media_name - name of media.ext (e.g., image.png)
            media_type - 'image' or 'video'
            
        
        
        -----        
        Project data API
        request:
        {"operation": "UploadMedia",
        "mime_type": "video/mp4",
        "media_name": "SocialInteractionDemoVideo.mp4",
        "media_size": "37201059"}
        
        response:
        {"operation": "UploadMedia",
        "status_code": 0}
        '''   
        
        # File extension
        file_ext = media_name.split('.')[1]
        
        # Check that media format is supported
        assert(len([i for i in ['bmp','jpeg', 'png','gif','mp4', 'x-msvideo'] \
                    if file_ext == i]) > 0), "File type not supported"
        
        print('upload_media')
        
        return {
                "operation": "UploadMedia",
                "status_code": 0,
                "md5_checksum": "dummy_md5_checksum",
                "media_id": "dummy_media_id"
        }
          
    #%% TODO: remove?
    def upload_media_abort(self):
        ''' Project data API
        request:
        {"operation": "UploadMediaAbort"}
        
        response:
        {"operation": "UploadMediaAbort",
        "status_code": 0}
        '''    
                
        print('upload_media_abort')
        
        return {"operation": "UploadMediaAbort",
                "status_code": 0}
 
    #%%
    def add_aois_to_image(self, media_id, aoi_name, aoi_color, 
                  key_frame_vertices, key_frame_active = True, 
                  key_frame_seconds = 0, tag_name = '',
                  group_name='', merge_mode='replace_aois'):
        
        
        ''' Project data API
        request:
    
        {
        "operation": "AddAois",
        "media_id": "b3418558-b245-4185-9bdc-725fb6a23a88",
        "aois":  {'name:' : 'test',
                  'color' : 'AACC33',
                  'key_frames' : {'is_active' : true,
                                  'seconds' : 0,
                                  'vertices' : {{100, 100},
                                                {100, 200},
                                                {200, 200},
                                                {200, 100}}},
                  'tags': {'tag_name' : 'vehicle',
                          'group_name' : 'Brand'}},
        "merge_mode": "ReplaceAois"
        }  
                
        response:
        {
        "operation": "sendAois",
        "imported_aoi_count": 1
        }            
        ''' 
        
        print('add_aois_to_image')
        
        return {"operation": "sendAois",
                "imported_aoi_count": 1
                } 
    
   #%%
    def add_aois_to_video(self, media_id, aoi_name, aoi_color, 
                  key_frame_vertices, key_frame_active = True, 
                  key_frame_seconds = 0, tag_name = '',
                  group_name='', merge_mode='replace_aois'):
        ''' 
        
        Placeholder for adding AOIs to video. RTFM!
        
        Project data API
        request:
    
        {
        "operation": "AddAois",
        "media_id": "b3418558-b245-4185-9bdc-725fb6a23a88",
        "aois":  {'name:' : 'test',
                  'color' : 'AACC33',
                  'key_frames' : {'is_active' : true,
                                  'seconds' : 0,
                                  'vertices' : {{100, 100},
                                                {100, 200},
                                                {200, 200},
                                                {200, 100}}},
                  'tags': {'tag_name' : 'vehicle',
                          'group_name' : 'Brand'}},
        "merge_mode": "ReplaceAois"
        }  
                
        response:
        {
        "operation": "sendAois",
        "imported_aoi_count": 1
        }            
        '''     
    #%%
    def get_state(self):
        ''' External presenter API.
        Returns the current state of External Presenter workflow.
        
        request:
        {
        "operation": "GetState"
        }
        response:
        {
        "operation": "GetState",
        "status_code": 0,
        "state": "unmet"
        }        
        '''
		
        print('get_state')
        
        return {
                "operation": "GetState",
                "status_code": 0,
                "state": "ready"
                }

    #%%           
    def start_recording(self, recording_name, participant_id, screen_width,
                        screen_height, screen_latency=10000):
        ''' External presenter API     
        request:
        {"operation": "StartRecording",
        "recording_name": "Waterfall",
        "participant_id": "3E6C3751-389A-47F0-861D-7A09D744FC23",
        "screen_latency": 10000,
        "screen_width": 1024,
        "screen_height": 768}
        
        response:
        {"operation": "StartRecording",
        "status_code": 0,
        "recording_id": "4B7F3751-389A-47D0-861D-7A09D744FC12"}
        
        Note: Participant identifier (returned in
                                      AddParticipant operation
                                      response)
        '''
        
        print('start_recording')
        
        return {"operation": "StartRecording",
                "status_code": 0,
                "recording_id": "dummy_recording_id"}
    #%%           
    def stop_recording(self):
        ''' External presenter API     
        request:
        {"operation": "StopRecording"}
        
        response:
        {"operation": "StopRecording",
        "status_code": 0}
        '''   
        
        
        print('stop_recording')
        
        return {"operation": "StopRecording",
                "status_code": 0}
        
    #%%           
    def send_stimulus_event(self, recording_id, start_timestamp, 
                            media_id, media_position=None, background=None,
                            end_timestamp=None):
                            
        ''' External presenter API.
        Timestamps should be in microseconds
        
        request:
        {"operation": "SendStimulusEvent",
        "recording_id": "4B7F3751-389A-47D0-861D-7A09D744FC12",
        "start_timestamp": "123123123123213123",
        "end_timestamp": "321321321321321321",
        "media_id": "2A86C895-F8B6-4786-9C4B-CB889C3449F0",
        "media_position":  {"left": 10,
                            "top": 20,
                            "right": 850,
                            "bottom": 1100},
        "background": "AA91C0"}
        
        response:
        {"operation": "SendStimulusEvent",
        "status_code": 0}
        '''       
        
        print('send_stimulus_event')
        
    #%%           
    def send_custom_event(self):
        ''' External presenter API     
        request:
        {"operation": "SendCustomEvent",
        "recording_id": "4B7F3751-389A-47D0-861D-7A09D744FC12",
        "timestamp": "123123123123213123",
        "event_type": "ERROR_SEM_OWNER_DIED",
        "value": "105 (0x69)"}
        
        response:
        {"operation": "SendCustomEvent",
        "status_code": 0}
        
        '''         
        
        print('send_custom_event')
        
    #%%           
    def finalize_recording(self, recording_id):
        ''' Finalizes the recording and makes it ready for analysis in the 
        application.
        
        If validation of data was sent, Server responds with status code 201. 
        Once the root cause of a validation failure has been fixed, 
        the command can be executed again.        
        
        External presenter API.  
        request:
        {"operation": "FinalizeRecording",
        "recording_id": "4B7F3751-389A-47D0-861D-7A09D744FC12"}
        
        response:
        {"operation": "FinalizeRecording",
        "status_code": 0}
        
        Note: recording_id is returned from start_recording
        '''             
        
        print('finalize_recording')
                
        
    #%% 
    def disconnect(self):
        ''' Closes the websocket connection
        '''
        print('disconnect')
                
        
        


__version__ = "2.0.0rc10"

__title__ = "titta"
__description__ = "Titta is a toolbox for using eye trackers from Tobii with Python, specifically offering integration with PsychoPy."
__url__ = "https://github.com/marcus-nystrom/Titta"
__uri__ = __url__
__doc__ = __description__ + " <" + __url__ + ">"

__author__ = "Marcus Nystrom"
__email__ = "marcus.nystrom@humlab.lu.se"

__license__ = "Creative Commons Attribution 4.0 (CC BY 4.0)"
__copyright__ = "Copyright (c) 2019-2023 " + __author__



from setuptools import setup
from distutils.util import convert_path

main_ns = {}
ver_path = convert_path('titta/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

setup(
    name=main_ns['__title__'],
    version=main_ns['__version__'],
    author=main_ns['__author__'],
    author_email=main_ns['__email__'],
    description=main_ns['__description__'],
    long_description=open('README.md').read(),
    long_description_content_type = 'text/markdown',
    url=main_ns['__url__'],
    project_urls={
        "Source Code": main_ns['__url__'],
    },
    license=main_ns['__license__'],

    packages=['titta'],
    python_requires=">=3.8",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
    ],
    install_requires=['numpy','pandas','h5py','TittaPy','websocket_client'],
)



# -*- coding: utf-8 -*-
"""
Created on Thu Jun 01 14:11:57 2017

@author: Marcus

    """
import pandas as pd
import copy
import sys
import warnings
import json
import pickle
from pathlib import Path
import os
import numpy as np
import h5py
import time
import TittaPy
import titta
from titta import helpers_tobii as helpers

# Suppress FutureWarning
warnings.simplefilter(action='ignore', category=FutureWarning)

# test if psychopy available
HAS_PSYCHOPY = False
try:
    import psychopy
    from psychopy import visual, core, event
except:
    pass
else:
    HAS_PSYCHOPY = True


#%%
class myTobii(object):
    """
    A Class to communicate with Tobii eye trackers
    """

    def __init__(self, settings):
        '''
        Constructs an instance of the Titta interface, with specified settings
        acquited through the call (Titta.get_defaults)
        '''

        self.settings = settings

        # Remove empty spaces (if any are given)
        self.settings.TRACKER_ADDRESS = self.settings.TRACKER_ADDRESS.replace(" ", "")


    #%%
    def init(self):
        ''' Apply settings and check capabilities
        '''

        # Start the eye tracker logger
        TittaPy.start_logging()

        # Update the number of calibration point (if changed by the user)
        if self.settings.N_CAL_TARGETS == 13:
            self.settings.CAL_POS_TOBII = self.settings.CAL_TARGETS[:]
        elif self.settings.N_CAL_TARGETS == 9:
            self.settings.CAL_POS_TOBII = self.settings.CAL_TARGETS[[0, 1, 2, 5, 6, 7, 10, 11, 12]]
        elif self.settings.N_CAL_TARGETS == 5:
            self.settings.CAL_POS_TOBII = self.settings.CAL_TARGETS[[0, 2, 6, 10, 12]]
        elif self.settings.N_CAL_TARGETS == 3:
            self.settings.CAL_POS_TOBII = self.settings.CAL_TARGETS[[0, 6, 12]]
        elif self.settings.N_CAL_TARGETS == 1:
            self.settings.CAL_POS_TOBII = self.settings.CAL_TARGETS[[6]]
        elif self.settings.N_CAL_TARGETS == 0:
            self.settings.CAL_POS_TOBII = np.array([])
        else:
            raise ValueError('Unvalid number of calibration points')

        # If no tracker address is given, find one automatically
        if len(self.settings.TRACKER_ADDRESS) == 0:

            # Sometimes you have to try a few times before it finds and eye tracker
            tracker_found = False
            k = 0
            while not tracker_found and k < 3:

                # if the tracker doesn't connect, try four times to reconnect
                ets = TittaPy.find_all_eye_trackers()
                for et in ets:

                    # Check if the desired eye tracker is found
                    if et['model'] == self.settings.eye_tracker_name:
                        self.settings.TRACKER_ADDRESS = et['address']
                        tracker_found = True
                        break

                time.sleep(1)
                k += 1

            if not tracker_found:
                raise Exception('The desired eye tracker not found. \
                                These are available: ' + ' '.join([str(m['model']) for m in ets]))

        if self.settings.PACING_INTERVAL < 0.8:
            raise Exception('Calibration pacing interval must be \
                            larger or equal to 0.8 s')

        # Initiate the EyeTracker class with a specific address
        self.buffer = TittaPy.EyeTracker(self.settings.TRACKER_ADDRESS)

        # Always include eye openness data in gaze stream
        if TittaPy.capability.has_eye_openness_data in self.buffer.capabilities:
            self.buffer.set_include_eye_openness_in_gaze(True)

        # Store timestamped messages in a list
        self.msg_container = []

        self.user_position_guide_data = None
        self.all_validation_results = []

        self.clock = core.Clock()

        # header / column names
        self.header = ['device_time_stamp',
                 'system_time_stamp',
                 'left_gaze_point_on_display_area_x',
                 'left_gaze_point_on_display_area_y',
                 'left_gaze_point_in_user_coordinate_system_x',
                 'left_gaze_point_in_user_coordinate_system_y',
                 'left_gaze_point_in_user_coordinate_system_z',
                 'left_gaze_origin_in_trackbox_coordinate_system_x',
                 'left_gaze_origin_in_trackbox_coordinate_system_y',
                 'left_gaze_origin_in_trackbox_coordinate_system_z',
                 'left_gaze_origin_in_user_coordinate_system_x',
                 'left_gaze_origin_in_user_coordinate_system_y',
                 'left_gaze_origin_in_user_coordinate_system_z',
                 'left_pupil_diameter',
                 'left_pupil_validity',
                 'left_gaze_origin_validity',
                 'left_gaze_point_validity',
                 'left_eyeopenness'
                 'right_gaze_point_on_display_area_x',
                 'right_gaze_point_on_display_area_y',
                 'right_gaze_point_in_user_coordinate_system_x',
                 'right_gaze_point_in_user_coordinate_system_y',
                 'right_gaze_point_in_user_coordinate_system_z',
                 'right_gaze_origin_in_trackbox_coordinate_system_x',
                 'right_gaze_origin_in_trackbox_coordinate_system_y',
                 'right_gaze_origin_in_trackbox_coordinate_system_z',
                 'right_gaze_origin_in_user_coordinate_system_x',
                 'right_gaze_origin_in_user_coordinate_system_y',
                 'right_gaze_origin_in_user_coordinate_system_z',
                 'right_pupil_diameter',
                 'right_pupil_validity',
                 'right_gaze_origin_validity',
                 'right_gaze_point_validity']

        # Only the tobii pro spectrum and the tobii pro fusion can record eye images
        # Never record eye images for other models, override defaults
        if not self.buffer.has_stream('eye_image'):
            if self.settings.RECORD_EYE_IMAGES_DURING_CALIBRATION:
                print('Warning: This eye tracker does not support eye images')
                self.settings.RECORD_EYE_IMAGES_DURING_CALIBRATION = False

        # Check and set sampling frequency if not correct
        Fs = self.get_sample_rate()
        if Fs != self.settings.SAMPLING_RATE:
            self.set_sample_rate(self.settings.SAMPLING_RATE)

        # Assert that the selected tracking mode is supported (human, primate)
        assert np.any([tracking_mode == self.settings.TRACKING_MODE \
                        for tracking_mode in self.buffer.supported_modes]), \
            "The given tracking mode is not supported. \
            Supported modes are: {}".format(self.buffer.supported_modes)

        try:
            print('Current tracking mode: {}'.format(self.buffer.tracking_mode))
            self.buffer.tracking_mode = self.settings.TRACKING_MODE
        except NameError:
            print('Tracking mode not found: {}'.format(self.settings.TRACKING_MODE))

        # Modify location and number of eye images for fusion
        if self.settings.eye_tracker_name == 'Tobii Pro Fusion':
            # Parameters for eye images (default values are for Spectrum)
            self.settings.graphics.EYE_IMAGE_SIZE = (0.25, 0.25)
            self.settings.graphics.EYE_IMAGE_POS_R = (0.55, -0.4)
            self.settings.graphics.EYE_IMAGE_POS_L = (-0.25, -0.4)
            self.settings.graphics.EYE_IMAGE_POS_R_1 = (0.25, -0.4) # Used for the two additional fusion images
            self.settings.graphics.EYE_IMAGE_POS_L_1 = (-0.55, -0.4)

           #%%
    def _create_calibration_buttons(self):
        '''Creates click buttons that are used during calibration
        If another window object, win_operator, is given, the experiment
        will be run in dual screen mode.

        '''

        # Find out ratio aspect of the stimulus screen
        screen_res = self.win.size
        ratio = screen_res[0] / float(screen_res[1])

        # Shown during setup to check gaze in corners
        self.POS_CAL_CHECK_DOTS = [[-0.45 * ratio, -0.45], [0.45 * ratio, -0.45],
                              [-0.45 * ratio, 0.45], [0.45 * ratio, 0.45]]

        # Text object to draw text (on buttons)
        instruction_text = visual.TextStim(self.win,text='',wrapWidth = 1,
                                           height = self.settings.graphics.TEXT_SIZE,
                                           color = self.settings.graphics.TEXT_COLOR,
                                           units='norm')
        self.instruction_text = instruction_text
        instruction_text_op = visual.TextStim(self.win_temp,text='',wrapWidth = 1,
                                              height = self.settings.graphics.TEXT_SIZE,
                                              color = self.settings.graphics.TEXT_COLOR,
                                              units='norm')
        self.instruction_text_op = instruction_text_op


        # Setup stimuli for drawing calibration / validation targets
        self.cal_dot = helpers.MyDot2(self.win, units='pix',
                                     outer_diameter=self.settings.graphics.TARGET_SIZE,
                                     inner_diameter=self.settings.graphics.TARGET_SIZE_INNER)

        # Click buttons
        self.calibrate_button = visual.Rect(self.win_temp, width= self.settings.graphics.WIDTH_CAL_BUTTON,
                                            height=self.settings.graphics.HEIGHT_CAL_BUTTON,
                                        units='norm', fillColor=self.settings.graphics.COLOR_CAL_BUTTON,
                                        pos=self.settings.graphics.POS_CAL_BUTTON)
        self.calibrate_button_text = visual.TextStim(self.win_temp, text=self.settings.graphics.CAL_BUTTON_TEXT,
                                                     height=self.settings.graphics.TEXT_SIZE,
                                                     color = self.settings.graphics.TEXT_COLOR,
                                                     units='norm',
                                                     pos=self.settings.graphics.POS_CAL_BUTTON)

        self.recalibrate_button = visual.Rect(self.win_temp, width= self.settings.graphics.WIDTH_RECAL_BUTTON,
                                            height=self.settings.graphics.HEIGHT_RECAL_BUTTON,
                                        units='norm', fillColor=self.settings.graphics.COLOR_RECAL_BUTTON,
                                        pos=self.settings.graphics.POS_RECAL_BUTTON)
        self.recalibrate_button_text = visual.TextStim(self.win_temp, text=self.settings.graphics.RECAL_BUTTON_TEXT,
                                                     height=self.settings.graphics.TEXT_SIZE,
                                                     color = self.settings.graphics.TEXT_COLOR,
                                                     units='norm',
                                                     pos=self.settings.graphics.POS_RECAL_BUTTON)

        self.revalidate_button = visual.Rect(self.win_temp, width= self.settings.graphics.WIDTH_REVAL_BUTTON,
                                            height=self.settings.graphics.HEIGHT_REVAL_BUTTON,
                                        units='norm', fillColor=self.settings.graphics.COLOR_REVAL_BUTTON,
                                        pos=self.settings.graphics.POS_REVAL_BUTTON)
        self.revalidate_button_text = visual.TextStim(self.win_temp, text=self.settings.graphics.REVAL_BUTTON_TEXT,
                                                     height=self.settings.graphics.TEXT_SIZE,
                                                     color = self.settings.graphics.TEXT_COLOR,
                                                     units='norm',
                                                     pos=self.settings.graphics.POS_REVAL_BUTTON)

        self.setup_button = visual.Rect(self.win_temp, width= self.settings.graphics.WIDTH_SETUP_BUTTON,
                                        height=self.settings.graphics.HEIGHT_SETUP_BUTTON,
                                        units='norm', fillColor=self.settings.graphics.COLOR_SETUP_BUTTON,
                                        pos=self.settings.graphics.POS_SETUP_BUTTON)
        self.setup_button_text = visual.TextStim(self.win_temp, text=self.settings.graphics.SETUP_BUTTON_TEXT,
                                                 height=self.settings.graphics.TEXT_SIZE, units='norm',
                                                 color = self.settings.graphics.TEXT_COLOR,
                                                 pos=self.settings.graphics.POS_SETUP_BUTTON)

        self.accept_button = visual.Rect(self.win_temp, width= self.settings.graphics.WIDTH_ACCEPT_BUTTON,
                                        height=self.settings.graphics.HEIGHT_ACCEPT_BUTTON,
                                        units='norm', fillColor=self.settings.graphics.COLOR_ACCEPT_BUTTON,
                                        pos=self.settings.graphics.POS_ACCEPT_BUTTON)
        self.accept_button_text = visual.TextStim(self.win_temp, text=self.settings.graphics.ACCEPT_BUTTON_TEXT,
                                                  height=self.settings.graphics.TEXT_SIZE,
                                                  color = self.settings.graphics.TEXT_COLOR,
                                                  units='norm',
                                                  pos=self.settings.graphics.POS_ACCEPT_BUTTON)

        self.back_button = visual.Rect(self.win_temp, width= self.settings.graphics.WIDTH_BACK_BUTTON,
                                        height=self.settings.graphics.HEIGHT_BACK_BUTTON,
                                        units='norm', fillColor=self.settings.graphics.COLOR_BACK_BUTTON,
                                        pos=self.settings.graphics.POS_BACK_BUTTON)
        self.back_button_text = visual.TextStim(self.win_temp, text=self.settings.graphics.BACK_BUTTON_TEXT,
                                                height=self.settings.graphics.TEXT_SIZE,
                                                color = self.settings.graphics.TEXT_COLOR,
                                                units='norm',
                                                pos=self.settings.graphics.POS_BACK_BUTTON)

        self.gaze_button = visual.Rect(self.win_temp, width= self.settings.graphics.WIDTH_GAZE_BUTTON,
                                        height=self.settings.graphics.HEIGHT_GAZE_BUTTON,
                                        units='norm', fillColor=self.settings.graphics.COLOR_GAZE_BUTTON,
                                        pos=self.settings.graphics.POS_GAZE_BUTTON)
        self.gaze_button_text = visual.TextStim(self.win_temp, text=self.settings.graphics.GAZE_BUTTON_TEXT,
                                                height=self.settings.graphics.TEXT_SIZE, units='norm',
                                                color = self.settings.graphics.TEXT_COLOR,
                                                pos=self.settings.graphics.POS_GAZE_BUTTON)

        self.calibration_image = visual.Rect(self.win_temp, width= self.settings.graphics.WIDTH_CAL_IMAGE_BUTTON,
                                        height=self.settings.graphics.HEIGHT_CAL_IMAGE_BUTTON,
                                        units='norm', fillColor=self.settings.graphics.COLOR_CAL_IMAGE_BUTTON,
                                        pos=self.settings.graphics.POS_CAL_IMAGE_BUTTON)

        self.calibration_image_text = visual.TextStim(self.win_temp, text=self.settings.graphics.CAL_IMAGE_BUTTON_TEXT,
                                                height=self.settings.graphics.TEXT_SIZE, units='norm',
                                                color = self.settings.graphics.TEXT_COLOR,
                                                pos=self.settings.graphics.POS_CAL_IMAGE_BUTTON)

        # Dots for the setup screen
        self.setup_dot = helpers.MyDot2(self.win, units='height',
                                     outer_diameter=self.settings.graphics.SETUP_DOT_OUTER_DIAMETER,
                                     inner_diameter=self.settings.graphics.SETUP_DOT_INNER_DIAMETER)


        # Dot for showing et data
        self.et_sample_l = visual.Circle(self.win_temp, radius = self.settings.graphics.ET_SAMPLE_RADIUS,
                                         fillColor = 'red', units='pix')
        self.et_sample_r = visual.Circle(self.win_temp, radius = self.settings.graphics.ET_SAMPLE_RADIUS,
                                         fillColor = 'blue', units='pix')

        if self.win_operator:
            self.raw_et_sample_l = visual.Circle(self.win_temp, radius = 0.01,
                                             fillColor = 'red', units='norm')
            self.raw_et_sample_r = visual.Circle(self.win_temp, radius = 0.01,
                                             fillColor = 'blue', units='norm')
            self.current_point = helpers.MyDot2(self.win_temp, units='norm',
                                         outer_diameter=0.05,
                                         inner_diameter=0.02)

        # Show images (eye image, validation resutls)
        self.eye_image_stim_l = visual.ImageStim(self.win_temp, units='norm',
                                                   size=self.settings.graphics.EYE_IMAGE_SIZE,
                                                   pos=self.settings.graphics.EYE_IMAGE_POS_L,
                                                   image=np.zeros((512, 512)))
        self.eye_image_stim_r = visual.ImageStim(self.win_temp, units='norm',
                                                   size=self.settings.graphics.EYE_IMAGE_SIZE,
                                                   pos=self.settings.graphics.EYE_IMAGE_POS_R,
                                                   image=np.zeros((512, 512)))
        self.eye_image_stim_l_1 = visual.ImageStim(self.win_temp, units='norm',
                                                   size=self.settings.graphics.EYE_IMAGE_SIZE,
                                                   pos=self.settings.graphics.EYE_IMAGE_POS_L_1,
                                                   image=np.zeros((512, 512)))
        self.eye_image_stim_r_1 = visual.ImageStim(self.win_temp, units='norm',
                                                   size=self.settings.graphics.EYE_IMAGE_SIZE,
                                                   pos=self.settings.graphics.EYE_IMAGE_POS_R_1,
                                                   image=np.zeros((512, 512)))

        # Accuracy image
        self.accuracy_image = visual.ImageStim(self.win_temp, image=None,units='norm', size=(2,2),
                                          pos=(0, 0))

    #%%
    def start_recording(self,   gaze=False,
                                time_sync=False,
                                eye_image=False,
                                notifications=False,
                                external_signal=False,
                                positioning=False):
        ''' Starts recording streams
        See bug re. positioning stream (ordering of starting and stopping streams)
        https://github.com/dcnieho/Titta/blob/master/tests/positioningStreamBug.m
        '''

        if gaze and self.buffer.has_stream('gaze'):
            self.buffer.start('gaze')
        if time_sync and self.buffer.has_stream('time_sync'):
            self.buffer.start('time_sync')
        if eye_image and self.buffer.has_stream('eye_image'):
            self.buffer.start('eye_image')
        if notifications and self.buffer.has_stream('notification'):
            self.buffer.start('notification')
        if external_signal and self.buffer.has_stream('external_signal'):
            self.buffer.start('external_signal')
        if positioning and self.buffer.has_stream('positioning'):
            self.buffer.start('positioning')


    #%%
    def stop_recording(self,    gaze=False,
                                time_sync=False,
                                eye_image=False,
                                notifications=False,
                                external_signal=False,
                                positioning=False):
        ''' Stops recording streams
        '''
        if positioning and self.buffer.has_stream('positioning'):
            self.buffer.stop('positioning')
        if gaze and self.buffer.has_stream('gaze'):
            self.buffer.stop('gaze')
        if time_sync and self.buffer.has_stream('time_sync'):
            self.buffer.stop('time_sync')
        if eye_image and self.buffer.has_stream('eye_image'):
            self.buffer.stop('eye_image')
        if notifications and self.buffer.has_stream('notification'):
            self.buffer.stop('notification')
        if external_signal and self.buffer.has_stream('external_signal'):
            self.buffer.stop('external_signal')

    #%%
    def calibrate(self, win, win_operator=None,
                  eye='both', calibration_number = 'second'):
        ''' Enters the calibration screen and gives the user control
        over what happens
        Args:
            win - psychopy window object
            win_operator - psychopy window object for the operator screen (optional)
            eye - 'left', 'right', or 'both' - default
            calibration_number - 'first' or 'second'. If 'first', don't exit
                                 calibration mode until 'second' calibration
                                 has finished

        '''

        # Only the Tobii Pro Spectrum currently support calibrating one eye at the time
        if not TittaPy.capability.can_do_monocular_calibration in self.buffer.capabilities:
            assert eye=='both', 'Monocular calibrations available only in Tobii Pro Spectrum'

        # Generate coordinates for calibration in PsychoPy and
        # Tobii coordinate system
        if self.settings.N_CAL_TARGETS > 0:
            CAL_POS_PIX = helpers.tobii2pix(self.settings.CAL_POS_TOBII, win)
            self.CAL_POS = np.hstack((self.settings.CAL_POS_TOBII, CAL_POS_PIX))

        VAL_POS_PIX = helpers.tobii2pix(self.settings.VAL_POS_TOBII, win)
        self.VAL_POS = np.hstack((self.settings.VAL_POS_TOBII, VAL_POS_PIX))

        # Create a temp variable for screen object
        if win_operator:
            win_temp = win_operator
        else:
            win_temp = win

        self.eye = eye
        self.win = win
        self.win_operator = win_operator
        self.win_temp = win_temp
        self.mouse = event.Mouse(win=win_temp)
        self.mouse.setVisible(0)

        # Used to keep track on monocular calibrations
        self.calibration_number = calibration_number

        # Create click buttons used during calibration
        self._create_calibration_buttons()

        # Animated calibration?
        if self.settings.ANIMATE_CALIBRATION:

            # Define your calibration target
            target = helpers.MyDot2(self.win, units='pix',
                                     outer_diameter=self.settings.graphics.TARGET_SIZE,
                                     inner_diameter=self.settings.graphics.TARGET_SIZE_INNER)
            self.animator = helpers.AnimatedCalibrationDisplay(self.win, target, 'animate_point')

        # Main control loop
        action = 'setup'
        self.deviations = [] # List to store validation accuracies
        self.buffer.start('gaze')
        self.buffer.start('time_sync')

        calibration_started = False

        # Select action
        while True:
            self.action = action
            if 'setup' in action:
                action = self._check_head_position()
            elif 'cal' in action:

                # Enter calibration mode
                if not calibration_started:
                    if self.eye == 'both':
                        # Enter calibration mode
                        if self.settings.DEBUG:
                            print('enter_calibration_mode')
                        self.buffer.enter_calibration_mode(False) # doMonocular=False
                        res = self._wait_for_action_complete()
                    else:
                        if self.calibration_number == 'first':
                            self.buffer.enter_calibration_mode(True) # doMonocular=True
                            res = self._wait_for_action_complete()
                        if self.settings.DEBUG:
                            print('enter_calibration_mode (monocular')

                # Default to last calibration when a new
                # Calibration is run
                self.selected_calibration = len(self.deviations) + 1

                # Run calibration only if number of targets exceeds 0
                if self.settings.N_CAL_TARGETS > 0:
                    action = self._run_calibration()
                else:
                    self.final_cal_position = ()
                    action = 'val'

                calibration_started = True
            elif 'val' in action:
                action = self._run_validation()
            elif 'res' in action:
                action = self._show_validation_screen()
            elif 'done'in action:
                if self.settings.DEBUG:
                    print('calibration completed')

                # Leave calibration mode if binocular calibration or if second
                # bi-monocular calibration
                if calibration_started:
                    if self.eye == 'both':
                        self.buffer.leave_calibration_mode(False)
                        res = self._wait_for_action_complete()

                    else:
                        if self.calibration_number == 'second':
                            self.buffer.leave_calibration_mode(True)
                            res = self._wait_for_action_complete()

                # Break out of loop
                break
            elif 'quit' in action:
                if calibration_started:
                    if self.eye == 'both':
                        self.buffer.leave_calibration_mode(False)
                    else:
                        self.buffer.leave_calibration_mode(True)

                    res = self._wait_for_action_complete()

                self.buffer.stop('gaze')
                self.buffer.stop('time_sync')
                self.win.close()
                if self.win_operator:
                    self.win_operator.close()
                core.quit()
            else:
                pass


            core.wait(0.01)

        self.buffer.stop('gaze')
        self.buffer.stop('time_sync')

        # Save all calibrations (appending 'used' means that the calibration was used)
        for i, devs in enumerate(self.deviations):
            if i == self.selected_calibration - 1:
                self.all_validation_results.append(devs + ['used'])
            else:
                self.all_validation_results.append(devs + ['not used'])

    #%%
    def set_dummy_mode(self):
        # Set current class to Dummy class
        from titta import Tobii_dummy
        self.__class__ = Tobii_dummy.Connect
        self.__class__.__init__(self)

    #%%
    def calibration_history(self):
        '''
        Get calibration history
        Validation results are stored in a list, where each list
        entry contains the accuracy values in degrees from
        [0] - left eye x
        [1] - left eye y
        [2] - right eye x
        [3] - right eye y

        The last entry [4] tells whether the calibration was used (1)
        or not used (0)

        So, if you want to extract the accuracy of the right eye y for the
        third attempted calibration, type
        all_validation_results[2][]

        '''
        return self.all_validation_results
    #%%
    def _draw_gaze(self):
        '''Shows gaze contingent cursor'''

        d = self.buffer.peek_N('gaze',1)

        lx = d['left_gaze_point_on_display_area_x'][0]
        ly = d['left_gaze_point_on_display_area_y'][0]
        rx = d['right_gaze_point_on_display_area_x'][0]
        ry = d['right_gaze_point_on_display_area_y'][0]

        self.et_sample_l.pos = helpers.tobii2pix(np.array([[lx, ly]]),
                                                 self.win_temp)
        self.et_sample_l.draw()
        self.et_sample_r.pos = helpers.tobii2pix(np.array([[rx, ry]]),
                                                 self.win_temp)
        self.et_sample_r.draw()

   #%%
    def _check_head_position(self):
        ''' Check to make sure that the head is in the center of the track box
        Two circles are shown; one static (representing the center of the headbox)
        and one variable representing the head position. The head is in the center
        of the trackbox the circles overlap.

        Four dot are shown in the corners. Ask the participants to fixate them
        to make sure there are not problems in the corners of the screen
        '''

        t0  = self.get_system_time_stamp()

        if self.settings.DEBUG:
            print('start_head_positioning')
        # Clear all events in mouse buffer
        event.clearEvents()

        self.mouse.setVisible(1)

        # Start user positioning guide
        self.buffer.start('positioning')

        # core.wait(0.5)

        if not self.buffer.has_stream('gaze'):
            self.win.close()
            raise Exception('Eye tracker switched on?')

        # Initiate parameters of head class (shown on participant screen)
        et_head = helpers.EThead(self.win)

        # Initiate parameters of head class (shown on operator screen)
        if self.win_operator:
            et_head_op = helpers.EThead(self.win_temp)

        show_eye_images = False
        image_button_pressed = False
        while True:

            # Draw four dots in the corners
            for i in self.POS_CAL_CHECK_DOTS:
                self.setup_dot.set_pos(i)
                self.setup_dot.draw()

            # Draw buttons, one to calibrate
            self.calibrate_button.draw()
            self.calibrate_button_text.draw()

            # Get keys
            k = event.getKeys()

            # Draw setup button information and check whether is was selected
            if self.buffer.has_stream('eye_image'):
                self.setup_button.draw()
                self.setup_button_text.draw()
                if self.settings.graphics.SETUP_BUTTON in k or (self.mouse.isPressedIn(self.setup_button, buttons=[0]) and not image_button_pressed):
                    # Toggle visibility of eye images
                    show_eye_images = not show_eye_images
                    image_button_pressed = True

             # Get position of eyes in track box (wait for valid samples)
            sample = self.buffer.peek_N('gaze',1)
            while len(sample['left_pupil_diameter']) == 0:
                sample = self.buffer.peek_N('gaze',1)

            sample_user_position = self.buffer.peek_N('positioning',1)
            while len(sample_user_position['left_user_position_x']) == 0:
                sample_user_position = self.buffer.peek_N('positioning',1)

            # Draw et head on participant screen
            et_head.update(sample, sample_user_position, eye=self.eye)
            et_head.draw()

            # Draw et head on operator screen
            if self.win_operator:
                et_head_op.update(sample, sample_user_position, eye=self.eye)
                et_head_op.draw()

            # Draw instruction
            self.instruction_text.pos = (0, 0.8)
            self.instruction_text.text = 'Position yourself such that the two circles overlap.'
            self.instruction_text.draw()

            # Get and draw distance information
            l_pos = sample['left_gaze_origin_in_user_coordinates_z'][0]
            r_pos = sample['right_gaze_origin_in_user_coordinates_z'][0]

            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    self.instruction_text_op.pos = (0, 0.9)
                    self.instruction_text_op.text = ' '.join(['Distance:',
                                                               str(int(np.nanmean([l_pos, r_pos])/10.0)), 'cm'])
                    self.instruction_text_op.draw()
            except:
                pass

            # Show eye images if requested
            if show_eye_images:
                # Start streaming of eye images
                if self.buffer.has_stream('eye_image') and not self.buffer.is_recording('eye_image'):
                    self.buffer.start('eye_image')

                self._draw_eye_image()
            else:
                if self.buffer.has_stream('eye_image') and self.buffer.is_recording('eye_image'):
                    self.buffer.stop('eye_image')

            # check whether someone left clicked any of the buttons or pressed 'space'
            if self.settings.graphics.CAL_BUTTON in k or self.mouse.isPressedIn(self.calibrate_button, buttons=[0]):
                action = 'cal'
                break

            if 'escape' in k:
                action = 'quit'
                break

            # Option to skip calibration (no button for this?)
            if 'return' in k:
                action = 'done'
                break

            # Wait for release of show gaze_button
            if not np.any(self.mouse.getPressed()):
                image_button_pressed = False

            # Update the screen
            self.win.flip()
            if self.win_operator:
                self.win_temp.flip()

        self.mouse.setVisible(0)

        # Stop user position guide
        self.buffer.stop('positioning')

        # Clear information from all available streams
        t1 = self.get_system_time_stamp()
        self.buffer.clear_time_range('gaze', t0, t1)

        if self.buffer.has_stream('eye_image'):
            self.buffer.clear_time_range('eye_image', t0, t1)

        # Clear the screen
        self.win.flip()
        if self.win_operator:
            self.win_temp.flip()

        if self.settings.DEBUG:
            print('stop_head_positioning')

        return action

    #%% _
    def _draw_operator_screen(self, pos, sample):
        ''' Draws what the operator sees in dual screen mode

        Args:
            bg - background (either image with calibration points or validation points)
            pos - current position of calibration / validation dot
            sample - dict with eye tracker sample
        '''

        # Draw calibration / validation dots
        # self.bg.draw()

        # Draw current target (red cicle)
        self.current_point.set_pos(helpers.tobii2norm(np.expand_dims(pos, axis=0)))
        self.current_point.draw()

        # Draw data for the left and right eyes
        self.raw_et_sample_l.pos = helpers.tobii2norm(np.expand_dims((sample['left_gaze_point_on_display_area_x'][0],
                                                                      sample['left_gaze_point_on_display_area_y'][0]),
                                                                      axis=0))
        self.raw_et_sample_r.pos = helpers.tobii2norm(np.expand_dims((sample['right_gaze_point_on_display_area_x'][0],
                                                                      sample['right_gaze_point_on_display_area_y'][0]),
                                                                      axis=0))
        self.raw_et_sample_l.draw()
        self.raw_et_sample_r.draw()

        # Draw eye images for the left and right eyes
        self._draw_eye_image()

    #%%
    def _wait_for_action_complete(self):
        ''' Wait for an action to finish (and see if i succeded)

        returns:
            information about the completed action
            res-  e.g.,
            print(res.work_item.action)
            print(res.work_item.coordinates)
            print(res.status_string)
        '''

        res = None
        while res==None:
            res = self.buffer.calibration_retrieve_result()

        return res

    #%%
    def _run_calibration(self):
        ''' Calibrates the eye tracker
        Args:   target_pos: N x 4 array with target positions
                first two columns are Tobii coords (x, y) and the two last
                are position in degrees (PsychoPy coord system)
                pval: pacing interval (how long each dot is shown)
        '''


        self.win.flip()
        if self.win_operator:
            self.win_operator.flip()

        paval=self.settings.PACING_INTERVAL

        # Optional start recording of eye images
        if self.settings.RECORD_EYE_IMAGES_DURING_CALIBRATION or self.win_operator:
            if self.buffer.has_stream('eye_image') and not self.buffer.is_recording('eye_image'):
                self.buffer.start('eye_image')

        self.send_message('Calibration_start')

        action = 'setup'

        # Go through the targets one by one
        cal_pos=self.CAL_POS
        np.random.shuffle(cal_pos)
        t0 = self.clock.getTime()
        i = 0
        pos_old = [0, 0]
        animation_state = 'static'
        tick = 0
        autopace = self.settings.AUTO_PACE
        while True:

            k = event.getKeys()

            # Get current calibration target position
            pos = cal_pos[i, 2:] # Psychopoy and Tobii have different coordinate systems
            tobii_data = cal_pos[i, :2]

            # Animate calibration dots or show static dots?
            if self.settings.ANIMATE_CALIBRATION:
                if animation_state == 'move':
                    move_completed = self.animator.move_point(pos_old, (pos[0], pos[1]), tick)
                    if move_completed:
                        animation_state = 'static' # or move
                        tick = 0
                        t0 = self.clock.getTime()
                else:
                    self.animator.animate_target(0, (pos[0], pos[1]), tick)
            else:
                self.cal_dot.set_pos(pos)
                self.cal_dot.draw()

            if self.win_operator:
                self._draw_operator_screen(tobii_data,
                                           self.buffer.peek_N('gaze', 1))
                self.win_temp.flip()

            self.win.flip()

            # Send a message at the onset (first frame) of a new dot
            # Always happens when tick == 0
            if tick == 0 and animation_state == 'static':
                self.send_message('calibration point {} at position {} {}'.format(i, pos[0], pos[1]))

            # Time to switch to a new point
            if autopace == 2 or 'space' in k:

                # Switch to fully automatic pacing once first point is accepted
                if self.settings.AUTO_PACE == 1:
                    autopace = 2

                if (self.clock.getTime() - t0) > paval:

                    pos_old = pos[:]
                    i += 1

                    # Collect some calibration data
                    if self.eye == 'both':
                        self.buffer.calibration_collect_data(tobii_data)
                    else:
                        self.buffer.calibration_collect_data(tobii_data, self.eye)

                    res = self._wait_for_action_complete()

                    if self.settings.DEBUG:
                        print(res['work_item']['action'])
                        print(res['work_item']['coordinates'])
                        print(res['status_string'])

                    if i == self.settings.N_CAL_TARGETS:
                        self.final_cal_position = copy.deepcopy(pos)
                        break

                    t0 = self.clock.getTime()

                    tick = 0
                    animation_state = 'move'

            # Abort calibration if esc is pressed (and go back to setup screen)
            if 'escape' in k:
                return action

            # 'r' aborts and restarts calibration
            if 'r' in k:
                action = 'cal'
                return action

            tick += 1

        # Apply the calibration
        self.buffer.calibration_compute_and_apply()
        res = self._wait_for_action_complete()

        if self.settings.DEBUG:
            print(res['work_item']['action'])
            print(res['status_string'])
            print(res['calibration_result'])


        # Accept of redo the calibration?
        if res['status'] == 0 and len(res['calibration_result']['points']) > 0:
            action = 'val'
            cal_data = self._generate_calibration_image(res)
        else:
            self.instruction_text.text = 'Calibration unsuccessful.'
            self.instruction_text.draw()
            self.win.flip()
            core.wait(1)

        self.send_message('Calibration_stop')

        # Stop recording of eye images unless validation is next
        if 'val' not in action:
            if self.settings.RECORD_EYE_IMAGES_DURING_CALIBRATION or self.win_operator:
                if self.buffer.has_stream('eye_image') and self.buffer.is_recording('eye_image'):
                    self.buffer.stop('eye_image')

        return action


    #%%
    def _generate_calibration_image(self, calibration_result): # TODO: display of gaze data on image does not seem correct
        ''' Generates visual representation of calibration results
        Plot only those who are valid and used?
        '''

        cal_data = []   # where calibration deviations are stored
        xys_left = []   # container for gaze data (left eye)
        xys_right = []  # container for gaze data (right eye)

        # Display the results to the user (loop over each calibration point)
        for p in calibration_result['calibration_result']['points']:

            # Draw calibration dots
            x_dot = p['position_on_display_area_x']
            y_dot = p['position_on_display_area_y']
            self.cal_dot.fillColor = 'white'
            xy_dot = helpers.tobii2pix(np.array([[x_dot, y_dot]]),
                                       self.win)
            self.cal_dot.set_pos(xy_dot) # Tobii and psychopy have different coord systems
            self.cal_dot.draw()

            if self.eye == 'both' or self.eye == 'left':
                # Save gaze data for left eye to list
                x = p['samples_left_position_on_display_area_x']
                y = p['samples_left_position_on_display_area_y']
                xy_sample = helpers.tobii2pix(np.array([x, y]),
                                              self.win) # Tobii and psychopy have different coord systems
                for xy in xy_sample.T:
                    xys_left.append([xy[0], xy[1]])
                    cal_data.append([x_dot, y_dot, xy[0], xy[1], 'left'])

            if self.eye == 'both' or self.eye == 'right':
                 # Save gaze data for right eye to list
                 x = p['samples_right_position_on_display_area_x']
                 y = p['samples_right_position_on_display_area_y']
                 xy_sample = helpers.tobii2pix(np.array([x, y]),
                                               self.win) # Tobii and psychopy have different coord systems
                 for xy in xy_sample.T:
                     xys_right.append([xy[0], xy[1]])
                     cal_data.append([x_dot, y_dot, xy[0], xy[1], 'right'])

        samples = visual.ElementArrayStim(self.win,
                                          sizes=self.settings.graphics.ET_SAMPLE_RADIUS,
                                          fieldSize=(100, 100),
                                          nElements=np.max([len(xys_left),
                                                            len(xys_right)]),
                                          elementTex=None,
                                          elementMask='circle',
                                          units='pix')

        # Draw calibration gaze data samples for left eye and right eyes
        if self.eye == 'both' or self.eye == 'left':
            samples.xys = np.array(xys_left)
            samples.colors = 'red'
            samples.draw()

        if self.eye == 'both' or self.eye == 'right':
            samples.xys = np.array(xys_right)
            samples.colors = 'blue'
            samples.draw()

        # Save validation results as image
        # nCalibrations = len(self.deviations) + 1
        fname = 'calibration_image' + str(self.selected_calibration)+'.png'
        self.win.getMovieFrame(buffer='back')
        self.win.saveMovieFrames(fname)

        # Clear the back buffer without flipping the window
        self.win.clearBuffer()

        return cal_data

    #%%
    def _save_calibration(self, filename):
        ''' Save the calibration to a .bin file
        Args:
            filename: test
        Returns:
            -
        '''
        self.buffer.calibration_get_data()
        res = self._wait_for_action_complete()
        calibration_data = res['calibration_data']

        if self.settings.DEBUG:
            print(res['work_item']['action'])
            print(res['status_string'])

        # None is returned on empty calibration.
        if calibration_data != None:
            if self.settings.DEBUG:
                print("Saving calibration to file {} for eye tracker with serial number {}.".format(filename, self.buffer.serial_number))

            with open(filename + '.p', 'wb') as handle:
                pickle.dump(calibration_data, handle)
        else:
            if self.settings.DEBUG:
                print("No calibration available for eye tracker with serial number {0}.".format(self.buffer.serial_number))
    #%%
    def _load_calibration(self, filename):
        ''' Loads the calibration to a .bin file
        Args:
            filename: test
        Returns:
            -
        '''
        # Read the calibration from file.
        with open(filename + '.p', 'rb') as handle:
            calibration_data = pickle.load(handle)

        # Don't apply empty calibrations.
        if len(calibration_data) > 0:
            if self.settings.DEBUG:
                print("Applying calibration {} on eye tracker with serial number {}.".format(filename, self.buffer.serial_number))
            self.buffer.calibration_apply_data(calibration_data)
            res = self._wait_for_action_complete()

            if self.settings.DEBUG:
                print(res['work_item']['action'])
                print(res['status_string'])

    #%%
    def _run_validation(self):
        ''' Runs a validation to check the accuracy of the calibration
        '''

        target_pos = self.VAL_POS
        paval = self.settings.PACING_INTERVAL

        np.random.shuffle(target_pos)

        # Set this to true if you want all data from the validation to
        # end up in the main data file (default, False)
        self.send_message('Validation_start')

        action = 'setup'
        # Go through the targets one by one
        self.clock.reset()
        i = 0

        if self.settings.N_CAL_TARGETS == 0:
            pos_old = [0, 0]
        else:
            pos_old = self.final_cal_position

        validation_data = []
        xy_pos = []
        animation_state = 'move'
        tick = 0
        while True:

            k = event.getKeys()

            # Get current calibration target position
            pos = target_pos[i, 2:] # Psychopoy and Tobii have different coordinate systems

            # Animate calibration dots or show static dots?
            if self.settings.ANIMATE_CALIBRATION:
                if animation_state == 'move':
                    move_completed = self.animator.move_point(pos_old, (pos[0], pos[1]), tick)
                    if move_completed:
                        animation_state = 'static' # or move
                        tick = 0
                        self.clock.reset()
                else:
                    self.animator.animate_target(0, (pos[0], pos[1]), tick)
            else:
                animation_state == 'static'
                self.cal_dot.set_pos(pos)
                self.cal_dot.draw()

            if self.win_operator:
                self._draw_operator_screen(target_pos[i, :2], self.buffer.peek_N('gaze',1))
                self.win_temp.flip()

            self.win.flip()

            # Send a message at the onset (first frame) of a new dot
            # Always happens when tick == 0
            # animation_state 'static' means that the dot has completed the
            # movement to a new location.
            if tick == 0 and animation_state == 'static':
                self.send_message('validation point {} at position {} {}'.format(i, pos[0], pos[1]))

            # # Start collecting validation data for a point 500 ms after its onset
            # if (self.clock.getTime() >= 0.500 and self.clock.getTime() < 0.800) and not buffer_started:
            #     self.start_ring_buffer(sample_buffer_length=300)
            #     buffer_started = True

            # Get the last 300 ms of data from the buffer
            if self.clock.getTime() >= 0.800:
                sample = self.buffer.peek_N('gaze',
                                             int(self.buffer.frequency * 0.300))

            # Time to switch to a new point
            if self.settings.AUTO_PACE > 0 or 'space' in k:
                if self.clock.getTime() > paval:

                    # Collect validation data from this point
                    validation_data.append(sample)

                    # Save data as xy list (Change so it's saved to a np.array directly)
                    for ii in range(len(sample['system_time_stamp'])):
                        xy_pos.append([sample['system_time_stamp'][ii],
                                       sample['left_gaze_point_on_display_area_x'][ii],
                                       sample['left_gaze_point_on_display_area_y'][ii],
                                       sample['right_gaze_point_on_display_area_x'][ii],
                                       sample['right_gaze_point_on_display_area_y'][ii],
                                       pos[0], pos[1]])
                    pos_old = pos[:]

                    self.clock.reset()
                    tick = 0
                    animation_state = 'move'

                    i += 1

                    if i > (np.shape(target_pos)[0] - 1):
                        break

            # Abort calibration if esc is pressed (and go back to setup screen)
            if 'escape' in k:
                return action

            # 'r' aborts and restarts calibration
            if 'r' in k:
                action = 'cal'
                return action

            tick += 1


        ##############
        self.send_message('Validation_end')

        self.win.flip()

        # Convert data from Tobii coord system to PsychoPy coordinates
        gaze_pos = np.array(xy_pos)
        gaze_pos[:, 1:3] = helpers.tobii2pix(gaze_pos[:, 1:3], self.win)
        gaze_pos[:, 3:5] = helpers.tobii2pix(gaze_pos[:, 3:5], self.win)

        # Compute data quality per validation point
        deviation_l, rms_l, sd_l, data_loss_l = self._compute_data_quality(validation_data, target_pos[:, :2], 'left')
        deviation_r, rms_r, sd_r, data_loss_r = self._compute_data_quality(validation_data, target_pos[:, :2], 'right')

        if self.eye == 'left':
            deviation_r = np.nan
            rms_r = np.nan
            sd_r = np.nan
            data_loss_r = np.nan

        if self.eye == 'right':
            deviation_l = np.nan
            rms_l = np.nan
            sd_l = np.nan
            data_loss_l = np.nan

        # FIXME: C:\git\Titta\titta\Tobii.py: RuntimeWarning: Mean of empty slice
        data_quality_values = [np.nanmean(deviation_l), np.nanmean(deviation_r),
                                np.nanmean(rms_l), np.nanmean(rms_r),
                                np.nanmean(sd_l), np.nanmean(sd_r),
                                np.nanmean(data_loss_l), np.nanmean(data_loss_r)]

        self.deviations.insert(self.selected_calibration - 1,
                               data_quality_values)

        self.send_message('validation data quality Dev_L: {:.2f}, Dev_R: {:.2f} RMS_L: {:.2f}, RMS_R: {:.2f}, SD_L: {:.2f}, SD_R: {:.2f}, LOSS_L: {:.2f}, LOSS_R: {:.2f}' \
                          .format(data_quality_values[0],
                                  data_quality_values[1],
                                  data_quality_values[2],
                                  data_quality_values[3],
                                  data_quality_values[4],
                                  data_quality_values[5],
                                  data_quality_values[6],
                                  data_quality_values[7]))


        self._generate_validation_image(target_pos[:, 2:], gaze_pos[:, :-2])
        action = 'res' # Show validation results

        # Stop recording of eye images
        if self.settings.RECORD_EYE_IMAGES_DURING_CALIBRATION or self.win_operator:
            if self.buffer.has_stream('eye_image') and self.buffer.is_recording('eye_image'):
                self.buffer.stop('eye_image')

        return action

    #%%
    def _generate_validation_image(self, dot_positions, gaze_positions):
        ''' Generates visual representation of validation results

        Args:
            dot_positions - n x 2 list with x, y location of validation dots
            gaze_positions - n x 5 array, t, lx, ly, rx, ry

        All values should be in the current screen units
        '''

        # Used to draw many dots
        samples = visual.ElementArrayStim(self.win,
                                          sizes=self.settings.graphics.ET_SAMPLE_RADIUS,
                                          fieldSize = (100, 100),  #self.settings.SAMPLE_DOT_SIZE,
                                          nElements = gaze_positions.shape[0],
                                          elementTex=None, elementMask='circle', units='pix')

        # Show all dots...
        for p in dot_positions:
            self.cal_dot.set_pos(p )
            self.cal_dot.draw()

        #... and collected validation samples for left...
        if self.eye == 'both' or self.eye == 'left':
            samples.xys = gaze_positions[:, 1:3]
            samples.colors = 'red'
            samples.draw()

        #... and right eye
        if self.eye == 'both' or self.eye == 'right':
            samples.xys = gaze_positions[:, 3:5]
            samples.colors = 'blue'
            samples.draw()

        # Save validation results as image
        # nCalibrations = len(self.deviations)
        fname = 'validation_image' + str(self.selected_calibration) + '.png'
        self.win.getMovieFrame(buffer='back')
        self.win.saveMovieFrames(fname)

        # Clear the back buffer without flipping the window
        self.win.clearBuffer()

        # Delete samples
        del samples

    #%%
    def _adcs2ucs(self, v):
        ''' Convert 2D point from Active Display Coordinate System (ADCS) to
        3D point in Tobii User Coordinate System (USC)
        Args:
            v - (x, y) in ADCS, e.g, (0.54, 0.1), norm [0 -> 1]

        Returns
            u - (x, y, z) in UCS in mm
        '''

        display_area = self.buffer.display_area

#        print("Bottom Left: {0}".format(display_area.bottom_left))
#        print("Bottom Right: {0}".format(display_area.bottom_right))
#        print("Height: {0}".format(display_area.height))
#        print("Top Left: {0}".format(display_area.top_left))
#        print("Top Right: {0}".format(display_area.top_right))
#        print("Width: {0}".format(display_area.width))

        dx = (np.array(display_area['top_right']) - np.array(display_area['top_left'])) * v[0]
        dy = (np.array(display_area['bottom_left']) - np.array(display_area['top_left'])) * v[1]

        u = np.array(display_area['top_left']) + dx + dy

        return u
    #%%
    def _compute_data_quality(self, validation_data, val_point_positions, eye):
        ''' Computes data quality (deviation, rms) per validation point

        Args:
            validation_data - list with validation data per validation point
                validation_data[k] - dict with keys (see self.header), contains, e.g.,
                len(sample['system_time_stamp']
            val_point_positions - list with [x, y] pos of validation point
                                    in ADCS
            eye - 'left' or 'right'
        '''
        # For each point
        deviation_per_point = []
        rms_per_point = []
        sd_per_point = []
        data_loss_point = []
        for p, point_data in enumerate(validation_data):

            val_point_position = self._adcs2ucs(val_point_positions[p])

            # For each sample (in a validation point)
            deviation_per_sample = []
            angle_between_samples = []
            gaze_vector_old = 0
            n_invalid_samples = 0
            for i in range(len(point_data['system_time_stamp'])):
                gaze_vector = (point_data[f"{eye}_gaze_point_in_user_coordinates_x"][i] -
                               point_data[f"{eye}_gaze_origin_in_user_coordinates_x"][i],
                               point_data[f"{eye}_gaze_point_in_user_coordinates_y"][i] -
                               point_data[f"{eye}_gaze_origin_in_user_coordinates_y"][i],
                               point_data[f"{eye}_gaze_point_in_user_coordinates_z"][i] -
                               point_data[f"{eye}_gaze_origin_in_user_coordinates_z"][i])

                if np.any(np.isnan(gaze_vector)):
                    n_invalid_samples += 1

                eye_to_valpoint_vector = (val_point_position[0] -
                               point_data[f"{eye}_gaze_origin_in_user_coordinates_x"][i],
                               val_point_position[1] -
                               point_data[f"{eye}_gaze_origin_in_user_coordinates_y"][i],
                               val_point_position[2] -
                               point_data[f"{eye}_gaze_origin_in_user_coordinates_z"][i])

                # Compute RMS (diff between consecutive samples) and deviation
                deviation_per_sample.append(helpers.angle_between(gaze_vector, eye_to_valpoint_vector))
                if i > 0:
                    angle_between_samples.append(helpers.angle_between(gaze_vector, gaze_vector_old))

                gaze_vector_old = gaze_vector

            # Compute averages per point
            deviation_per_point.append(np.rad2deg(np.nanmedian(deviation_per_sample)))
            rms_per_point.append(np.rad2deg(helpers.rms(angle_between_samples)))
            sd_per_point.append(np.rad2deg(helpers.sd(angle_between_samples)))
            data_loss_point.append(n_invalid_samples / float(i))

        return deviation_per_point, rms_per_point, sd_per_point, data_loss_point


    #%%
    def _show_validation_screen(self):
        ''' Shows validation image after a validation has been completed
        '''

        # Get timestamp
        t0_clear = self.get_system_time_stamp()

        # Center position of presented calibration values
        x_pos_res = 0.55
        y_pos_res = 0.2

        self.mouse.setVisible(1)

        # get (and save) validation screen image
        nCalibrations = len(self.deviations)
        self._save_calibration(str(nCalibrations))

        # Add image as texture
        fname = 'validation_image' + str(self.selected_calibration)+'.png'
        self.accuracy_image.image = fname
        show_validation_image = True    # Default is to show validation results,
                                        # not calibration results

        # information about data quality header
        header = ['Quality', 'Left eye', 'Right eye', 'Left eye', 'Right eye',
                   'Left eye', 'Right eye']
        header_colors = [self.settings.graphics.TEXT_COLOR,
                         'red', 'blue', 'red', 'blue', 'red', 'blue']
        x_pos= np.linspace(-0.35, 0.35, num = 7)

        # Prepare rects for buttons, button text, and accuracy values
        select_accuracy_rect = []
        select_rect_text = []
        accuracy_values = []#[0] * len(x_pos)] * nCalibrations
        # print(accuracy_values)

        y_pos = y_pos_res

        for i in range(nCalibrations):
            select_accuracy_rect.append(visual.Rect(self.win_temp, width= 0.15,
                                                height= 0.05,
                                                units='norm',
                                                pos = (x_pos_res, y_pos)))

            select_rect_text.append(visual.TextStim(self.win_temp,
                                                    text='Select',
                                                    wrapWidth = 1,
                                                    height = self.settings.graphics.TEXT_SIZE,
                                                    color = self.settings.graphics.TEXT_COLOR,
                                                    units='norm',
                                                    pos = (x_pos_res, y_pos)))

            # Then prepare the accuracy values for each calibration preceded
            # by Cal x (the calibration number)
            accuracy_values_j = []
            for j, x in enumerate(x_pos):
                if j > 0:
                    accuracy_values_j.append(visual.TextStim(self.win_temp,
                                                        text='{0:.2f}'.format(self.deviations[i][j - 1]),
                                                        wrapWidth = 1,
                                                        height = self.settings.graphics.TEXT_SIZE,
                                                        units='norm',
                                                        color = header_colors[j],
                                                        pos = (x, y_pos)))
                else:
                    accuracy_values_j.append(visual.TextStim(self.win_temp,
                                                        text='Cal' + str(i+1) + ':',
                                                        wrapWidth = 1,
                                                        height = self.settings.graphics.TEXT_SIZE,
                                                        units='norm',
                                                        color = header_colors[j],
                                                        pos = (x, y_pos)))
            accuracy_values.append(accuracy_values_j)
            y_pos -= 0.06

        # Prepare main header
        header_main_text = ['Accuracy (deg)', 'Precision (deg)', 'Data loss (%)']
        self.instruction_text.setColor([1, 1, 1], colorSpace='rgb')
        y_pos_main = y_pos_res + 0.13
        x_pos_main = np.linspace(-0.17, 0.30, num = 3)

        header_text_main = []
        for j, x in enumerate(x_pos_main):
            header_text_main.append(visual.TextStim(self.win_temp,text=header_main_text[j],
                                                wrapWidth = 1,
                                                height = self.settings.graphics.TEXT_SIZE,
                                                units='norm',
                                                pos = (x, y_pos_main),
                                                color = self.settings.graphics.TEXT_COLOR))
        # Prepare header
        header_text = []
        self.instruction_text.setColor([1, 1, 1], colorSpace='rgb')
        y_pos = y_pos_res
        for j, x in enumerate(x_pos):
            header_text.append(visual.TextStim(self.win_temp,text=header[j],
                                                wrapWidth = 1,
                                                height = self.settings.graphics.TEXT_SIZE,
                                                units='norm',
                                                pos = (x, y_pos_res + 0.06),
                                                color = header_colors[j]))


        # Wait for user input
        selection_done = False
        display_gaze = False
        gaze_button_pressed = False
        cal_image_button_pressed = False
        timing = []

        # If there's an operator screen, just show a message 'please wait...'
        # on the participant screen
        if self.win_operator:
            self.instruction_text.text = 'Please wait...'

        while not selection_done:

            t0 = self.clock.getTime()

            # Draw validation results image
            self.accuracy_image.draw()

            # Draw buttons (re-calibrate, accept and move on, show gaze)
            self.recalibrate_button.draw()
            self.recalibrate_button_text.draw()

            self.revalidate_button.draw()
            self.revalidate_button_text.draw()

            self.accept_button.draw()
            self.accept_button_text.draw()

            self.gaze_button.draw()
            self.gaze_button_text.draw()

            if self.settings.N_CAL_TARGETS > 0:
                self.calibration_image.draw()
                self.calibration_image_text.draw()

            # Draw headers
            [h.draw() for h in header_text_main]
            [hh.draw() for hh in header_text]

            # Draw accuracy/precision values and buttons to select a calibration
            for i in range(nCalibrations):

                # Highlight selected calibrations
                if i == self.selected_calibration - 1: # Calibration selected
                    select_accuracy_rect[i].fillColor = self.settings.graphics.blue_active
                    if nCalibrations > 1:
                        select_accuracy_rect[i].draw()
                        select_rect_text[i].draw()
                else:
                    select_accuracy_rect[i].fillColor = self.settings.graphics.blue
                    select_accuracy_rect[i].draw()
                    select_rect_text[i].draw()

                # Then draw the accuracy values for each calibration preceded
                # by Cal x (the calibration number)
                for j, x in enumerate(x_pos):
                    accuracy_values[i][j].draw()

            # Check if mouse is clicked to select a calibration
            for i, button in enumerate(select_accuracy_rect):
                if self.mouse.isPressedIn(button):
                    self._load_calibration(str(i + 1))  # Load the selected calibration
                    if show_validation_image:
                        fname = 'validation_image' + str(i + 1) + '.png'
                    else:
                        fname = 'calibration_image' + str(i + 1) + '.png'
                    self.accuracy_image.image = fname
                    self.selected_calibration = int(i + 1)
                    break

            # Check if key or button is pressed
            k = event.getKeys()
            if self.settings.graphics.RECAL_BUTTON in k or self.mouse.isPressedIn(self.recalibrate_button):
                action = 'setup'
                selection_done = True
            elif self.settings.graphics.REVAL_BUTTON in k or self.mouse.isPressedIn(self.revalidate_button):

                action = 'val'

                # Delete previous validation results, since we want to
                # re-validate
                self.deviations.pop(self.selected_calibration - 1)

                selection_done = True
            elif self.settings.graphics.ACCEPT_BUTTON in k or self.mouse.isPressedIn(self.accept_button):
                action = 'done'
                selection_done = True
            elif k:
                if k[0].isdigit():
                    if any([s for s in range(nCalibrations+1) if s == int(k[0])]):
                        self._load_calibration(k[0])  # Load the selected calibration
                        fname = 'validation_image' + str(k[0]) + '.png'
                        self.accuracy_image.image = fname
                        self.selected_calibration = int(k[0])

            elif 'escape' in k:
                action = 'quit'
                break

            # Toggle display gaze
            if self.settings.graphics.GAZE_BUTTON in k or (self.mouse.isPressedIn(self.gaze_button, buttons=[0]) and not gaze_button_pressed):
                display_gaze = not display_gaze
                gaze_button_pressed = True

            # Display gaze along with four dots in the corners
            if display_gaze:
                for i in self.POS_CAL_CHECK_DOTS:
                    self.setup_dot.set_pos(i)
                    self.setup_dot.draw()
                self._draw_gaze()

            # Show calibration image or validation image
            if self.settings.N_CAL_TARGETS > 0:
                if self.settings.graphics.CAL_IMAGE_BUTTON in k or (self.mouse.isPressedIn(self.calibration_image, buttons=[0])and not cal_image_button_pressed):

                    # Toggle the state of the button
                    show_validation_image = not show_validation_image

                    # If validation image show, switch to calibration and vice versa
                    if show_validation_image:
                        fname = 'validation_image' + str(self.selected_calibration)+'.png'
                        self.accuracy_image.image = fname
                    else:
                        fname = 'calibration_image' + str(self.selected_calibration)+'.png'
                        self.accuracy_image.image = fname

                    cal_image_button_pressed = True

            # Wait for release of show gaze_button
            if not np.any(self.mouse.getPressed()):
                gaze_button_pressed = False
                cal_image_button_pressed = False

            timing.append(self.clock.getTime() - t0)

            if self.win_operator:
                self.win_temp.flip()
                self.instruction_text.draw()

            self.win.flip()

        # Clear data that were recorded during the validation screen
        t1_clear = self.get_system_time_stamp()
        self.buffer.clear_time_range('gaze', t0_clear, t1_clear)

        if self.buffer.has_stream('eye_image'):
            self.buffer.clear_time_range('eye_image', t0_clear, t1_clear)

        # Clear screen and return
        self.instruction_text.setColor([1, 1, 1], colorSpace='rgb')
        self.win.flip()
        self.mouse.setVisible(0)


        return action

    #%%
    def system_info(self):
        ''' Returns information about system in dict
        '''

        info = {}
        info['serial_number']  = self.buffer.serial_number
        info['address']  = self.buffer.address
        info['model']  = self.buffer.model
        info['name']  = self.buffer.device_name
        info['firmware_version'] = self.buffer.firmware_version
        info['runtime_version'] = self.buffer.runtime_version
        info['tracking_mode']  = self.buffer.tracking_mode
        info['sampling_frequency']  = self.buffer.frequency

        try:
            info['track_box']  = self.buffer.track_box
        except:
            print('track box not supported by this eye tracker')

        info['display_area']  = self.buffer.display_area
        info['python_version'] = '.'.join([str(sys.version_info[0]),
                                           str(sys.version_info[1]),
                                           str(sys.version_info[2])])
        info['psychopy_version'] = psychopy.__version__
        info['TittaPy_version'] = TittaPy.__version__
        info['titta_version'] = titta.__version__
        # info['git_revision'] = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).decode('ascii').strip()

        return info

    #%%
    def get_system_time_stamp(self):
        ''' Get system time stamp
        '''

        return TittaPy.get_system_timestamp()

    #%%
    def send_message(self, msg, ts=None):
        ''' Sends a message to the data file
        '''

        if not ts:
            ts = self.get_system_time_stamp()
        self.msg_container.append([ts, msg])


    #%%
    def _draw_eye_image(self):
        ''' Draw left and right eye image
        '''

        if self.buffer.has_stream('eye_image'):

            # Grab the two most recent eye images from the buffer
            eye_images = self.buffer.peek_N('eye_image', 2, 'end')

            for i in range(len(eye_images['image'])):

                # Read eye image
                im_arr = eye_images['image'][i]

                # Convert to have values between -1 adn 1
                im_arr = np.fliplr(np.flipud((im_arr / float(im_arr.max()) * 2.0) - 1))

                # Always display image from camera_id == 0 on left side
                if eye_images['camera_id'][i] == 0:
                    if eye_images['region_id'][i] == 0:
                        self.eye_image_stim_l.image = im_arr
                    else:
                        self.eye_image_stim_r_1.image = im_arr
                else:
                    if eye_images['region_id'][i] == 0:
                        self.eye_image_stim_r.image = im_arr
                    else:
                        self.eye_image_stim_l_1.image = im_arr

            # Draw eye images
            self.eye_image_stim_l.draw()
            self.eye_image_stim_r.draw()

            # The Fusion has 4 eye images, while Spectrum only 2
            if self.settings.eye_tracker_name == 'Tobii Pro Fusion':
                self.eye_image_stim_l_1.draw()
                self.eye_image_stim_r_1.draw()

    #%%
    def set_sample_rate(self, Fs):
        '''Sets the sample rate
        '''

        # print([i for i in self.buffer.supported_frequencies], Fs)
        assert np.any([int(i) == Fs for i in self.buffer.supported_frequencies]), "Supported frequencies are: {}".format(self.buffer.supported_frequencies)

        self.buffer.frequency = int(Fs)
    #%%
    def get_sample_rate(self):
        '''Gets the sample rate
        '''
        return self.buffer.frequency

    #%%
    def save_data(self, filename=None, append_version=True):
        ''' Saves the data to HDF5 container
        If you want to read the data, see the 'resources' folder

        Args:
            filename - if a filename is given, it overrides the name stored in
            append_version : if file exists, it is appended with filename_1 etc.
                THis is to prevent file from being overwritten
        '''

        t0 = time.time()

        if filename:
            fname = filename
        else:
            fname = self.settings.FILENAME

        # Check if the filename already exists, if so append to name
        i = 1
        filename_ext = ''
        files = Path.cwd().glob('*.h5')
        while True:

            # Go through the files and look for a match
            filename_exists = False
            for f in files:
                f_temp = str(f).split(os.sep)[-1].split('.')[0]

                # if the file exists
                if fname + filename_ext == f_temp:
                    if not append_version:
                        print('Warning! Filename already exists. Will be overwritten.')
                    else: # append '_i to filename
                        filename_ext = '_' + str(i)
                        i += 1
                    filename_exists = True
                    break

            # If we've gone through all files without
            # a match, we ready!
            if not filename_exists:
                 break

         # Add the new extension the the filename
        fname = os.sep.join([fname + filename_ext])

        # Save gaze data to HDF5 container
        temp = self.buffer.consume_N('gaze',sys.maxsize)
        pd.DataFrame.from_dict(temp).to_hdf(fname + '.h5', key='gaze')

        # Save messages as HDF5 container
        df_msg = pd.DataFrame(self.msg_container,  columns=['system_time_stamp', 'msg'])
        df_msg.to_hdf(fname + '.h5', key='msg')

        # Save all other gaze streams in the same HDF5 container
        if self.buffer.has_stream('time_sync'):
            temp = self.buffer.consume_N('time_sync',sys.maxsize)
            pd.DataFrame.from_dict(temp).to_hdf(fname + '.h5', key='time_sync')
        if self.buffer.has_stream('external_signal'):
            temp = self.buffer.consume_N('external_signal',sys.maxsize)

            # Change external_signal_change_type type into list
            temp['change_type'] = [t.name for t in temp['change_type']]

            pd.DataFrame.from_dict(temp).to_hdf(fname + '.h5', key='external_signal')
        if self.buffer.has_stream('notification'): # TODO: can notifications be stored like this? Check format. Same with logs. Remove these fields before storing?
            '''
            your performance may suffer as PyTables will pickle object types that it cannot
            map directly to c-types [inferred_type->mixed,key->block2_values] [items->Index(['notification_type', 'display_area', 'errors_or_warnings'], dtype='object')]
            [t.name for t in temp['notification_type']]
            '''
            temp = self.buffer.consume_N('notification',sys.maxsize)

            # Change notification type into list
            temp['notification_type'] = [t.name for t in temp['notification_type']]

            # Change change all others to strings to prevent above warning
            # columns = ['errors_or_warnings', 'display_area', 'output_frequency']
            # df.loc[:,columns] = df[columns].applymap(str)
            temp['display_area'] = [repr(t) for t in temp['display_area']]
            temp['output_frequency'] = [repr(t) for t in temp['output_frequency']]
            temp['errors_or_warnings'] = [repr(t) for t in temp['errors_or_warnings']]

            pd.DataFrame.from_dict(temp).to_hdf(fname + '.h5', key='notification')

       # Save calibration history to HDF5 container
        df_cal = pd.DataFrame(self.calibration_history(),  columns=['offset_left_eye (deg)',
                                                                  'offset_right_eye (deg)',
                                                                  'RMS_S2S_left_eye (deg)',
                                                                  'RMS_S2S_right_eye (deg)',
                                                                  'SD_left_eye (deg)',
                                                                  'SD_right_eye (deg)',
                                                                  'Prop_data_loss_left_eye',
                                                                  'Prop_data_loss_right_eye',
                                                                  'Calibration used'])
        df_cal.to_hdf(fname + '.h5', key='calibration_history')

        # Save tracker/python version info as json
        with open(fname + '_info.json', "w") as outfile:
            json.dump(self.system_info(), outfile)

        # Save image stream in the same HDF5 container
        temp = self.buffer.consume_N('eye_image')
        if len(temp['image']) > 0:

            with h5py.File(fname + '.h5', 'a') as hf:
                grp = hf.create_group('eye_image')

                # Save each frame as a separate dataset in this group
                # To access later, use [i[:] for i in grp.values()]
                for k, im in enumerate(temp['image']):
                    grp.create_dataset(str(k), data=im)

            # # Remove the numpy image and save the rest
            del temp['image']
            temp['type'] = [t.name for t in temp['type']] # LIst of strings, e.g, "full_image", "cropped_image"
            pd.DataFrame.from_dict(temp).to_hdf(fname + '.h5', key='eye_metadata')

        # Clear data containers
        self.msg_container = []
        self.all_validation_results = []

        # Stop logging and Save data from the python wrapper
        TittaPy.stop_logging()
        l=TittaPy.get_log(True)  # True means the log is consumed. False (default) its only peeked.
        if len(l) > 0:
            d =  {}
            for key in list(l[0].keys()):
                d[key] = [i[key] for i in l]

        # Convert to prevent warning when saving to Hdf5 (see notifications above)
        d['level'] = [t.name for t in d['level']]
        d['source'] = [t.name for t in d['source']]

        # Save log file
        pd.DataFrame.from_dict(d).to_hdf(fname + '.h5', key='log')

        print(f'Took {time.time() - t0} s to save the data')





